#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# License under MIT

import subprocess
import collections
import os
import platform
import re
import sys
import time
import threading
import json
import warnings

from uiautomator import device as d
from uiautomator import Device as UiaDevice
import cv2
import aircv as ac

from . import base
from . import proto
from . import patch

from atx import consts
from atx import errors


FindPoint = collections.namedtuple('FindPoint', ['pos', 'confidence'])
log = base.getLogger('devsuit')

__dir__ = os.path.dirname(os.path.abspath(__file__))
__tmp__ = os.path.join(__dir__, '__cache__')

DISPLAY_RE = re.compile(
    '.*DisplayViewport{valid=true, .*orientation=(?P<orientation>\d+), .*deviceWidth=(?P<width>\d+), deviceHeight=(?P<height>\d+).*')

class Watcher(object):
    ACTION_TOUCH = 1

    def __init__(self):
        self._events = {}

    def on(self, image, action):
        self._events[image] = action

    def hook(self, screen, d):
        for (img, action) in self._events.items():
            if action == Watcher.ACTION_TOUCH:
                ret = d.exists(img, screen=screen)
                if ret is None:
                    continue
                if ret.confidence < 0.9:
                    print("Skip confidence:", ret.confidence)
                    continue
                print('trigger watch click')
                d.click(*ret.pos)


class AndroidDevice(UiaDevice):
    def __init__(self, serialno=None):
        super(AndroidDevice, self).__init__(serialno)
        self._serial = serialno
        self._uiauto = super(AndroidDevice, self)
        self._minicap_params = None
        self._watchers = {}

        self.screenshot_method = consts.SCREENSHOT_METHOD_UIAUTOMATOR
        # print 'DEVSUIT_SERIALNO:', phoneno
        # self.dev = dev
        # self.appname = appname
        # self._devtype = devtype
        # self._inside_depth = 0

        # # default image search extentension and 
        # self._image_exts = ['.jpg', '.png']
        # self._image_dirs = ['.', 'image']

        # self._rotation = None # 0,1,2,3
        # self._tmpdir = 'tmp'
        # self._click_timeout = 20.0 # if icon not found in this time, then panic
        # self._delay_after_click = 0.5 # when finished click, wait time
        # self._screen_resolution = None

        # self._snapshot_file = None
        # self._keep_capture = False # for func:keepScreen,releaseScreen
        # # self._logfile = logfile
        # self._loglock = threading.Lock()
        # self._operation_mark = False

        # self._image_match_method = 'auto'
        # self._threshold = 0.3 # for findImage

        # # if self._logfile:
        # #     logdir = os.path.dirname(logfile) or '.'
        # #     if not os.path.exists(logdir):
        # #         os.makedirs(logdir)
        # #     if os.path.exists(logfile):
        # #         backfile = logfile+'.'+time.strftime('%Y%m%d%H%M%S')
        # #         os.rename(logfile, backfile)

        # # Only for android phone method=<adb|screencap>
        # def _snapshot_method(method):
        #     if method and self._devtype == 'android':
        #         self.dev._snapshot_method = method

        # self._snapshot_method = _snapshot_method
        #-- end of func setting

    def sleep(self, secs):
        secs = int(secs)
        for i in reversed(range(secs)):
            sys.stdout.write('\r')
            sys.stdout.write("sleep %ds, left %2ds" % (secs, i+1))
            sys.stdout.flush()
            time.sleep(1)
        sys.stdout.write("\n")

    def add_watcher(self, w):
        id = time.time()
        self._watchers[id] = w
        return id

    def del_watcher(self, id):
        self._watchers.pop(id, None)

    def _tmp_filename(self, prefix='tmp-', ext='.png'):
        return '%s%s%s' %(prefix, time.time(), ext)

    def _get_minicap_params(self):
        """
        Used about 0.1s
        uiautomator d.info is now well working with device which has virtual menu.
        """
        w, h = (0, 0)
        for line in self.shell('dumpsys display').splitlines():
            m = DISPLAY_RE.search(line, 0)
            if not m:
                continue
            w = int(m.group('width'))
            h = int(m.group('height'))
            # o = int(m.group('orientation'))
            w, h = min(w, h), max(w, h)
            self._minicap_params = '{x}x{y}@{x}x{y}/{r}'.format(
                x=w, y=h, r=d.info['displayRotation']*90)
            break
        else:
            raise errors.BaseError('Fail to get display width and height')
        return self._minicap_params
        
    def _minicap(self):
        phone_tmp_file = '/data/local/tmp/'+self._tmp_filename(ext='.jpg')
        local_tmp_file = os.path.join(__tmp__, self._tmp_filename(ext='.jpg'))
        command = 'LD_LIBRARY_PATH=/data/local/tmp /data/local/tmp/minicap -P {} -s > {}'.format(
            self._get_minicap_params(), phone_tmp_file)
        self.shell(command)
        self.adbrun(['pull', phone_tmp_file, local_tmp_file])
        self.shell(['rm', phone_tmp_file])
        img = cv2.imread(local_tmp_file)
        os.remove(local_tmp_file)
        return img

    def screenshot(self, filename=None):
        """
        Take screen snapshot

        Args:
            filename: filename where save to, optional

        Returns:
            cv2 Image object

        Raises:
            TypeError
        """
        screen = None
        if self.screenshot_method == consts.SCREENSHOT_METHOD_UIAUTOMATOR:
            tmp_file = os.path.join(__tmp__, self._tmp_filename())
            self._uiauto.screenshot(tmp_file)
            screen = cv2.imread(tmp_file)
            os.remove(tmp_file)
        elif self.screenshot_method == consts.SCREENSHOT_METHOD_MINICAP:
            screen = self._minicap()
        else:
            raise TypeError('Invalid screenshot_method')

        if filename:
            save_dir = os.path.dirname(filename) or '.'
            if not os.path.exists(save_dir):
                os.makedirs(save_dir)
            cv2.imwrite(filename, screen)

        # handle watchers
        for w in self._watchers.values():
            w.hook(screen, self)

        return screen

    def adbrun(self, command):
        '''
        Run adb command, for example: adbrun(['pull', '/data/local/tmp/a.png'])

        Args:
            command: string or list of string

        Returns:
            command output
        '''
        cmds = ['adb']
        if self._serial:
            cmds.extend(['-s', self._serial])

        if isinstance(command, list) or isinstance(command, tuple):
            cmds.extend(list(command))
        else:
            cmds.append(command)
        output = subprocess.check_output(cmds, stderr=subprocess.STDOUT)
        return output.replace('\r\n', '\n')
        # os.system(subprocess.list2cmdline(cmds))

    def shell(self, command):
        '''
        Run adb shell command

        Args:
            command: string or list of string

        Returns:
            command output
        '''
        if isinstance(command, list) or isinstance(command, tuple):
            return self.adbrun(['shell'] + list(command))
        else:
            return self.adbrun(['shell'] + [command])

    def start_app(self, package_name):
        '''
        Start application

        Args:
            package_name: string like com.example.app1

        Returns:
            None
        '''
        self.shell(['monkey', '-p', package_name, '-c', 'android.intent.category.LAUNCHER', '1'])
    
    def stop_app(self, package_name, clear=False):
        '''
        Stop application

        Args:
            package_name: string like com.example.app1
            clear: bool, remove user data

        Returns:
            None
        '''
        if clear:
            self.shell(['pm', 'clear', package_name])
        else:
            self.shell(['am', 'force-stop', package_name])

    def takeSnapshot(self, filename):
        '''
        Deprecated, use screenshot instead.
        '''
        warnings.warn("deprecated, use snapshot instead", DeprecationWarning)
        return self.screenshot(filename)

    def _read_img(self, img):
        if isinstance(img, basestring):
            return ac.imread(img)
        # FIXME(ssx): need support other types
        return img

    def exists(self, img, screen=None):
        """Change if image exists

        Args:
            img: string or opencv image
            screen: opencv image, optional, if not None, screenshot will be called

        Returns:
            None or FindPoint
        """
        search_img = self._read_img(img)
        if screen is None:
            screen = self.screenshot()
        ret = ac.find_template(screen, search_img)
        if ret is None:
            return None
        position = ret['result']
        confidence = ret['confidence']
        log.info('match confidence: %s', confidence)
        return FindPoint(position, confidence)

    def touch_image(self, img, timeout=20.0, wait_change=False):
        """Simulate touch according image position

        Args:
            img: filename or an opencv image object
            timeout: float, if image not found during this time, ImageNotFoundError will raise.
            wait_change: wait until background image changed.
        Returns:
            None

        Raises:
            ImageNotFoundError: An error occured when img not found in current screen.
        """
        search_img = self._read_img(img)
        log.info('touch image: %s', img)
        start_time = time.time()
        found = False
        while time.time() - start_time < timeout:
            point = self.exists(search_img)
            if point is None:
                continue
            self._uiauto.click(*point.pos)
            found = True
            break

        # wait until click area not same
        if found and wait_change:
            start_time = time.time()
            while time.time()-start_time < timeout:
                screen_img = self.screenshot()
                ret = ac.find_template(screen_img, search_img)
                if ret is None:
                    break

        if not found:
            raise errors.ImageNotFoundError('Not found image %s' %(img,))

    # def __getattribute__(self, name):
    #     # print name
    #     v = object.__getattribute__(self, name)
    #     if isinstance(v, collections.Callable):
    #         objdict = object.__getattribute__(self, '__dict__')
    #         # print objdict

    #         def _wrapper(*args, **kwargs):
    #             objdict['_inside_depth'] += 1
    #             # log function call
    #             ret = v(*args, **kwargs)
    #             if objdict['_inside_depth'] == 1 and \
    #                 not v.__name__.startswith('_') and \
    #                     not v.__name__ == 'log':
    #                 self.log(proto.TAG_FUNCTION, dict(name=v.__name__, args=args, kwargs=kwargs))
    #             objdict['_inside_depth'] -= 1
    #             return ret
    #         return _wrapper
    #     return v

    # def _imfind(self, bgimg, search):
    #     method = self._image_match_method
    #     print 'match-method:', method
    #     imsrc, imsch = ac.imread(bgimg), ac.imread(search)
    #     if method == 'auto':
    #         point = ac.find(imsrc, imsch)
    #     elif method == 'template':
    #         res = ac.find_template(imsrc, imsch, self._threshold)
    #         if res:
    #             point, score = res
    #             print 'match result:', point, score
    #             return point
    #         return None

    #     elif method == 'sift':
    #         point = imtsift.find(search, bgimg)
    #     else:
    #         raise RuntimeError("Unknown image match method: %s" %(method))
    #     return point

    # def _imfindall(self, bgimg, search, maxcnt, sort):
    #     if not maxcnt:
    #         maxcnt = 0
    #     method = self._image_match_method
    #     imsrc, imsch = ac.imread(bgimg), ac.imread(search)
    #     if method == 'auto':
    #         points = ac.find_all(imsrc, imsch, maxcnt=5)
    #         # points = imtauto.locate_more_image_Template(search, bgimg, num=maxcnt)
    #     elif method == 'template':
    #         points = imttemplate.findall(search, bgimg, self._threshold, maxcnt=maxcnt)
    #     elif method == 'sift':
    #         points = imtsift.findall(search, bgimg, maxcnt=maxcnt)
    #     else:
    #         raise RuntimeError("Unknown image match method: %s" %(method))
    #     if sort:
    #         def cmpy((x0, y0), (x1, y1)):
    #             return y1<y0

    #         def cmpx((x0, y0), (x1, y1)):
    #             return x1<x1
    #         m = {'x': cmpx, 'y': cmpy}
    #         points.sort(cmp=m[sort])
    #     return points

    # def rotation(self):
    #     '''
    #     device orientation
    #     @return int
    #     '''
    #     # 通过globalSet设置的rotation
    #     if self._rotation:
    #         return self._rotation
    #     # 看dev是否有rotation方法
    #     if hasattr(self.dev, 'rotation'):
    #         return self.dev.rotation()
    #     # windows上的特殊处理
    #     if self._devtype == 'windows':
    #         return proto.ROTATION_0
    #     return proto.ROTATION_0

    # def _fix_point(self, (x, y)):
    #     w, h = self.shape() # in shape() the width always < height
    #     if self.rotation() % 2 == 1:
    #         w, h = h, w

    #     if isinstance(x, float) and x <= 1.0:
    #         x = int(w*x)
    #     if isinstance(y, float) and y <= 1.0:
    #         y = int(h*y)
    #     return (x, y)

    def _search_image(self, filename):
        ''' Search image in default path '''
        if isinstance(filename, unicode) and platform.system() == 'Windows':
            filename = filename.encode('gbk')
            #filename = filename.encode('utf-8')
        basename, ext = os.path.splitext(filename)
        exts = [ext] if ext else self._image_exts
        for folder in self._image_dirs:
            for ext in exts:
                fullpath = os.path.join(folder, basename+ext)
                if os.path.exists(fullpath):
                    return fullpath
        raise RuntimeError('Image file(%s) not found in %s' %(filename, self._image_dirs))

    def _save_screen(self, filename, random_name=True, tempdir=True):
        # use last snapshot file
        if self._snapshot_file and self._keep_capture:
            return self._snapshot_file

        if random_name:
            filename = base.random_name(filename)
        if tempdir:
            filename = os.path.join(self._tmpdir, filename)

        parent_dir = os.path.dirname(filename) or '.'
        if not os.path.exists(parent_dir):
            base.makedirs(parent_dir)

        # FIXME(ssx): don't save as file, better store in memory
        self.dev.snapshot(filename)

        if tempdir:
            self.log(proto.TAG_SNAPSHOT, dict(filename=filename))
        self._snapshot_file = filename
        return filename

    # def log(self, tag, message):
    #     if not self._logfile:
    #         return

    #     self._loglock.acquire()
    #     timestamp = time.time()
    #     try:
    #         dirname = os.path.dirname(self._logfile) or '.'
    #         if not os.path.exists(dirname):
    #             os.path.makedirs(dirname)
    #     except:
    #         pass
    #     with open(self._logfile, 'a') as file:
    #         data = dict(timestamp=int(timestamp), tag=tag, data=message)
    #         file.write(json.dumps(data) + '\n')
    #     self._loglock.release()

    def keepCapture(self):
        '''
        Use screen in memory
        '''
        self._keep_capture = True

    def releaseCapture(self):
        '''
        Donot use screen in memory (this is default behavior)
        '''
        self._keep_capture = False

    def globalSet(self, *args, **kwargs):
        '''
        app setting, be careful you should known what you are doing.
        @parma m(dict): eg:{"threshold": 0.3}
        '''
        if len(args) > 0:
            m = args[0]
            assert isinstance(m, dict)
        else:
            m = kwargs
        for k, v in m.items():
            key = '_'+k
            if hasattr(self, key):
                item = getattr(self, key)
                if callable(item):
                    item(v)
                else:
                    setattr(self, key, v)
            else:
                print 'not have such setting: %s' %(k)

    def globalGet(self, key):
        '''
        get app setting
        '''
        if hasattr(self, '_'+key):
            return getattr(self, '_'+key)
        return None

    def startApp(self, appname, activity):
        '''
        Start app
        '''
        self.dev.start_app(appname, activity)

    def stopApp(self, appname):
        '''
        Stop app
        '''
        self.dev.stop_app(appname)

    def find(self, imgfile):
        '''
        Find image position on screen

        @return (point founded or None if not found)
        '''
        filepath = self._search_image(imgfile)
        
        log.debug('Locate image path: %s', filepath)
        
        screen = self._save_screen('screen-{t}-XXXX.png'.format(t=time.strftime("%y%m%d%H%M%S")))
        if self._screen_resolution:
            # resize image
            ow, oh = self._screen_resolution # original
            cw, ch = self.shape() # current
            (ratew, rateh) = cw/float(ow), ch/float(oh)

            im = cv2.imread(filepath, cv2.IMREAD_UNCHANGED)

            nim = cv2.resize(im, (0, 0), fx=ratew, fy=rateh)
            new_name = base.random_name('resize-{t}-XXXX.png'.format(t=time.strftime("%y%m%d%H%M%S")))
            filepath = new_name = os.path.join(self._tmpdir, new_name)
            cv2.imwrite(new_name, nim)
            # im.resize((int(ratew*rw), int(rateh*rh))).save(new_name)
            # filepath = new_name
        pt = self._imfind(screen, filepath)
        return pt

    def mustFind(self, imgfile):
        ''' 
        Raise Error if image not found
        '''
        pt = self.find(imgfile)
        if not pt:
            raise RuntimeError("Image[%s] not found" %(imgfile))
        return pt

    def findall(self, imgfile, maxcnt=None, sort=None):
        '''
        Find multi positions that imgfile on screen

        @maxcnt (int): max number of object restricted.
        @sort (string): (None|x|y) x to sort with x, small in front, None to be origin order
        @return list point that found
        @warn not finished yet.
        '''
        filepath = self._search_image(imgfile)
        screen = self._save_screen('find-XXXXXXXX.png')
        pts = self._imfindall(screen, filepath, maxcnt, sort)
        return pts

    def safeWait(self, imgfile, seconds=20.0):
        '''
        Like wait, but don't raise RuntimeError

        return None when timeout
        return point if found
        '''
        warnings.warn("deprecated, use safe_wait instead", DeprecationWarning)
        self.safe_wait(imgfile, seconds)

    def safe_wait(self, img, seconds=20.0):
        '''
        Like wait, but don't raise RuntimeError

        return None when timeout
        return point if found
        '''
        warnings.warn("deprecated, use safe_wait instead", DeprecationWarning)
        try:
            return self.wait(img, seconds)
        except:
            return None        

    def wait(self, imgfile, timeout=20):
        '''
        Wait until some picture exists
        @return position when imgfile shows
        @raise RuntimeError if not found
        '''
        log.info('WAIT: %s', imgfile)
        start = time.time()
        while True:
            pt = self.find(imgfile)
            if pt:
                return pt
            if time.time()-start > timeout: 
                break
            time.sleep(1)
        raise RuntimeError('Wait timeout(%.2f)', float(timeout))

    # def exists(self, imgfile):
        # return True if self.find(imgfile) else False

    # def click(self, img_or_point, timeout=None, duration=None):
    #     '''
    #     Click function
    #     @param seconds: float (if time not exceed, it will retry and retry)
    #     '''
    #     if timeout is None:
    #         timeout = self._click_timeout
    #     log.info('CLICK %s, timeout=%.2fs, duration=%s', img_or_point, timeout, str(duration))
    #     point = self._val_to_point(img_or_point)
    #     if point:
    #         (x, y) = point
    #     else:
    #         (x, y) = self.wait(img_or_point, timeout=timeout)
    #     log.info('Click %s point: (%d, %d)', img_or_point, x, y)
    #     self.dev.touch(x, y, duration)
    #     log.debug('delay after click: %.2fs', self._delay_after_click)

    #     # FIXME(ssx): not tested
    #     if self._operation_mark:
    #         if self._snapshot_file and os.path.exists(self._snapshot_file):
    #             img = ac.imread(self._snapshot_file)
    #             ac.mark_point(img, (x, y))
    #             cv2.imwrite(self._snapshot_file, img)
    #         if self._devtype == 'android':
    #             self.dev.adbshell('am', 'broadcast', '-a', 'MP_POSITION', '--es', 'msg', '%d,%d' %(x, y))

    #     time.sleep(self._delay_after_click)

    def center(self):
        '''
        Center position
        '''
        w, h = self.shape()
        return w/2, h/2
    
    def clickIfExists(self, imgfile):
        '''
        Click when image file exists

        @return (True|False) if clicked
        '''
        log.info('CLICK IF EXISTS: %s' %(imgfile))
        pt = self.find(imgfile)
        if pt:
            log.debug('click for exists %s', imgfile)
            self.click(pt)
            return True
        else:
            log.debug('ignore for no exists %s', imgfile)
            return False

    def drag(self, fpt, tpt, duration=0.5):
        ''' 
        Drag from one place to another place

        @param fpt,tpt: filename or position
        @param duration: float (duration of the event in seconds)
        '''
        fpt = self._val_to_point(fpt)
        tpt = self._val_to_point(tpt)
        return self.dev.drag(fpt, tpt, duration)

    # def sleep(self, secs=1.0):
    #     '''
    #     Sleeps for the specified number of seconds

    #     @param secs: float (number of seconds)
    #     @return None
    #     '''
    #     log.debug('SLEEP %.2fs', secs)
    #     time.sleep(secs)

    def type(self, text):
        '''
        Input some text

        @param text: string (text want to type)
        '''
        self.dev.type(text)

    @patch.run_once
    def shape(self):
        '''
        Get device shape

        @return (width, height), width < height
        '''
        return sorted(self.dev.shape())

    def keyevent(self, event):
        '''
        Send keyevent (only support android and ios)

        @param event: string (one of MENU,BACK,HOME)
        @return nothing
        '''
        if hasattr(self.dev, 'keyevent'):
            return self.dev.keyevent(event)
        raise RuntimeError('keyevent not support')



'''
Copyright (C) 2012-2015  Diego Torres Milano
Created on Dec 1, 2012

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Diego Torres Milano
'''

__version__ = '10.0.1'

import sys
import warnings
# import string
# import datetime
if sys.executable:
    if 'monkeyrunner' in sys.executable:
        warnings.warn(
    '''

    You should use a 'python' interpreter, not 'monkeyrunner' for this module

    ''', RuntimeWarning)
import socket
import time
import re
import signal
import os
# import types
import platform

# from com.dtmilano.android.window import Window
# from com.dtmilano.android.common import _nd, _nh, _ns, obtainPxPy, obtainVxVy,\
#     obtainVwVh
# from com.dtmilano.android.adb.androidkeymap import KEY_MAP
from .androidkeymap import KEY_MAP

DEBUG = False
DEBUG_TOUCH = DEBUG and False
DEBUG_LOG = DEBUG and False
DEBUG_WINDOWS = DEBUG and False
DEBUG_COORDS = DEBUG and False

try:
    HOSTNAME = os.environ['ANDROID_ADB_SERVER_HOST']
except:
    HOSTNAME = 'localhost'

try:
    PORT = int(os.environ['ANDROID_ADB_SERVER_PORT'])
except KeyError:
    PORT = 5037

OKAY = 'OKAY'
FAIL = 'FAIL'

UP = 0
DOWN = 1
DOWN_AND_UP = 2

TIMEOUT = 15

WIFI_SERVICE = 'wifi'

# some device properties
VERSION_SDK_PROPERTY = 'ro.build.version.sdk'
VERSION_RELEASE_PROPERTY = 'ro.build.version.release'


class Device:
    @staticmethod
    def factory(_str):
        if DEBUG:
            print >> sys.stderr, "Device.factory(", _str, ")"
        values = _str.split(None, 2)
        if DEBUG:
            print >> sys.stderr, "values=", values
        return Device(*values)

    def __init__(self, serialno, status, qualifiers=None):
        self.serialno = serialno
        self.status = status
        self.qualifiers = qualifiers

    def __str__(self):
        return "<<<" + self.serialno + ", " + self.status + ", %s>>>" % self.qualifiers

class WifiManager:
    '''
    Simulates Android WifiManager.
    
    @see: http://developer.android.com/reference/android/net/wifi/WifiManager.html
    '''
    
    WIFI_STATE_DISABLING = 0
    WIFI_STATE_DISABLED = 1
    WIFI_STATE_ENABLING = 2
    WIFI_STATE_ENABLED = 3
    WIFI_STATE_UNKNOWN = 4
    
    WIFI_IS_ENABLED_RE = re.compile('Wi-Fi is enabled')
    WIFI_IS_DISABLED_RE = re.compile('Wi-Fi is disabled')
    
    def __init__(self, device):
        self.device = device
        
    def getWifiState(self):
        '''
        Gets the Wi-Fi enabled state.
        
        @return: One of WIFI_STATE_DISABLED, WIFI_STATE_DISABLING, WIFI_STATE_ENABLED, WIFI_STATE_ENABLING, WIFI_STATE_UNKNOWN
        '''
        
        result = self.device.shell('dumpsys wifi')
        if result:
            state = result.splitlines()[0]
            if self.WIFI_IS_ENABLED_RE.match(state):
                return self.WIFI_STATE_ENABLED
            elif self.WIFI_IS_DISABLED_RE.match(state):
                return self.WIFI_STATE_DISABLED
        print >> sys.stderr, "UNKNOWN WIFI STATE:", state
        return self.WIFI_STATE_UNKNOWN

class AdbClient:

    def __init__(self, serialno=None, hostname=HOSTNAME, port=PORT, settransport=True, reconnect=True, ignoreversioncheck=False):
        self.Log = AdbClient.__Log(self)
        
        self.serialno = serialno
        self.hostname = hostname
        self.port = port

        self.reconnect = reconnect
        self.__connect()

        self.checkVersion(ignoreversioncheck)

        self.build = {}
        ''' Build properties '''

        self.__displayInfo = None
        ''' Cached display info. Reset it to C{None} to force refetching display info '''

        self.display = {}
        ''' The map containing the device's physical display properties: width, height and density '''

        self.isTransportSet = False
        if settransport and serialno != None:
            self.__setTransport()
            self.build[VERSION_SDK_PROPERTY] = int(self.__getProp(VERSION_SDK_PROPERTY))
            self.initDisplayProperties()


    @staticmethod
    def setAlarm(timeout):
        osName = platform.system()
        if osName.startswith('Windows'):  # alarm is not implemented in Windows
            return
        if DEBUG:
            print >> sys.stderr, "setAlarm(%d)" % timeout
        signal.alarm(timeout)

    def setSerialno(self, serialno):
        if self.isTransportSet:
            raise ValueError("Transport is already set, serialno cannot be set once this is done.")
        self.serialno = serialno
        self.__setTransport()
        self.build[VERSION_SDK_PROPERTY] = int(self.__getProp(VERSION_SDK_PROPERTY))

    def setReconnect(self, val):
        self.reconnect = val

    def __connect(self):
        if DEBUG:
            print >> sys.stderr, "__connect()"
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(TIMEOUT)
        try:
            self.socket.connect((self.hostname, self.port))
        except socket.error, ex:
            raise RuntimeError("ERROR: Connecting to %s:%d: %s.\nIs adb running on your computer?" % (self.socket, self.port, ex))

    def close(self):
        if DEBUG:
            print >> sys.stderr, "Closing socket...", self.socket
        if self.socket:
            self.socket.close()

    def __del__(self):
        try:
            self.close()
        except:
            pass

    def __send(self, msg, checkok=True, reconnect=False):
        if DEBUG:
            print >> sys.stderr, "__send(%s, checkok=%s, reconnect=%s)" % (msg, checkok, reconnect)
        if not re.search('^host:', msg):
            if not self.isTransportSet:
                self.__setTransport()
        else:
            self.checkConnected()
        b = bytearray(msg, 'utf-8')
        self.socket.send('%04X%s' % (len(b), b))
        if checkok:
            self.__checkOk()
        if reconnect:
            if DEBUG:
                print >> sys.stderr, "    __send: reconnecting"
            self.__connect()
            self.__setTransport()

    def __receive(self, nob=None):
        if DEBUG:
            print >> sys.stderr, "__receive()"
        self.checkConnected()
        if nob is None:
            nob = int(self.socket.recv(4), 16)
        if DEBUG:
            print >> sys.stderr, "    __receive: receiving", nob, "bytes"
        recv = bytearray()
        nr = 0
        while nr < nob:
            chunk = self.socket.recv(min((nob - nr), 4096))
            recv.extend(chunk)
            nr += len(chunk)
        if DEBUG:
            print >> sys.stderr, "    __receive: returning len=", len(recv)
        return str(recv)

    def __checkOk(self):
        if DEBUG:
            print >> sys.stderr, "__checkOk()"
        self.checkConnected()
        self.setAlarm(TIMEOUT)
        recv = self.socket.recv(4)
        if DEBUG:
            print >> sys.stderr, "    __checkOk: recv=", repr(recv)
        try:
            if recv != OKAY:
                error = self.socket.recv(1024)
                if error.startswith('0049'):
                    raise RuntimeError("ERROR: This computer is unauthorized. Please check the confirmation dialog on your device.")
                else:
                    raise RuntimeError("ERROR: %s %s" % (repr(recv), error))
        finally:
            self.setAlarm(0)
        if DEBUG:
            print >> sys.stderr, "    __checkOk: returning True"
        return True

    def checkConnected(self):
        if DEBUG:
            print >> sys.stderr, "checkConnected()"
        if not self.socket:
            raise RuntimeError("ERROR: Not connected")
        if DEBUG:
            print >> sys.stderr, "    checkConnected: returning True"
        return True

    def checkVersion(self, ignoreversioncheck=False, reconnect=True):
        if DEBUG:
            print >> sys.stderr, "checkVersion(reconnect=%s)   ignoreversioncheck=%s" % (reconnect, ignoreversioncheck)
        self.__send('host:version', reconnect=False)
        # HACK: MSG_WAITALL not available on windows
        #version = self.socket.recv(8, socket.MSG_WAITALL)
        version = self.__readExactly(self.socket, 8)

        VALID_ADB_VERSIONS = ["00040020", "0004001f"]

        if not (version in VALID_ADB_VERSIONS) and not ignoreversioncheck:
            raise RuntimeError("ERROR: Incorrect ADB server version %s (expecting one of %s)" % (version, VALID_ADB_VERSIONS))
        if reconnect:
            self.__connect()

    def __setTransport(self):
        if DEBUG:
            print >> sys.stderr, "__setTransport()"
        if not self.serialno:
            raise ValueError("serialno not set, empty or None")
        self.checkConnected()
        serialnoRE = re.compile(self.serialno)
        found = False
        devices = self.getDevices()
        if len(devices) == 0:
            raise RuntimeError("ERROR: There are no connected devices")
        for device in devices:
            if serialnoRE.match(device.serialno):
                found = True
                break
        if not found:
            raise RuntimeError("ERROR: couldn't find device that matches '%s' in %s" % (self.serialno, devices))
        self.serialno = device.serialno
        msg = 'host:transport:%s' % self.serialno
        if DEBUG:
            print >> sys.stderr, "    __setTransport: msg=", msg
        self.__send(msg, reconnect=False)
        self.isTransportSet = True

    def __checkTransport(self):
        if not self.isTransportSet:
            raise RuntimeError("ERROR: Transport is not set")
    
    def __readExactly(self, sock, size):
        if DEBUG:
            print >> sys.stderr, "__readExactly(socket=%s, size=%d)" % (socket, size)
        _buffer = ''
        while len(_buffer) < size:
            data = sock.recv(size-len(_buffer))
            if not data:
                break
            _buffer+=data
        return _buffer

    def getDevices(self):
        if DEBUG:
            print >> sys.stderr, "getDevices()"
        self.__send('host:devices-l', checkok=False)
        try:
            self.__checkOk()
        except RuntimeError, ex:
            print >> sys.stderr, "**ERROR:", ex
            return None
        devices = []
        for line in self.__receive().splitlines():
            devices.append(Device.factory(line))
        self.__connect()
        return devices

    def shell(self, cmd=None):
        if DEBUG:
            print >> sys.stderr, "shell(cmd=%s)" % cmd
        self.__checkTransport()
        if cmd:
            self.__send('shell:%s' % cmd, checkok=True, reconnect=False)
            out = ''
            while True:
                _str = None
                try:
                    _str = self.socket.recv(4096)
                except Exception, ex:
                    print >> sys.stderr, "ERROR:", ex
                if not _str:
                    break
                out += _str
            if self.reconnect:
                if DEBUG:
                    print >> sys.stderr, "Reconnecting..."
                self.close()
                self.__connect()
                self.__setTransport()
            return out
        else:
            self.__send('shell:')
            # sin = self.socket.makefile("rw")
            # sout = self.socket.makefile("r")
            # return (sin, sin)
            sout = adbClient.socket.makefile("r")
            return sout

    def __getRestrictedScreen(self):
        ''' Gets C{mRestrictedScreen} values from dumpsys. This is a method to obtain display dimensions '''

        rsRE = re.compile('\s*mRestrictedScreen=\((?P<x>\d+),(?P<y>\d+)\) (?P<w>\d+)x(?P<h>\d+)')
        for line in self.shell('dumpsys window').splitlines():
            m = rsRE.match(line)
            if m:
                return m.groups()
        raise RuntimeError("Couldn't find mRestrictedScreen in 'dumpsys window'")

    def getDisplayInfo(self):
        self.__checkTransport()
        displayInfo = self.getLogicalDisplayInfo()
        if displayInfo:
            return displayInfo
        displayInfo = self.getPhysicalDisplayInfo()
        if displayInfo:
            return displayInfo
        raise RuntimeError("Couldn't find display info in 'wm size', 'dumpsys display' or 'dumpsys window'")

    def getLogicalDisplayInfo(self):
        '''
        Gets C{mDefaultViewport} and then C{deviceWidth} and C{deviceHeight} values from dumpsys.
        This is a method to obtain display logical dimensions and density
        '''

        self.__checkTransport()
        logicalDisplayRE = re.compile('.*DisplayViewport{valid=true, .*orientation=(?P<orientation>\d+), .*deviceWidth=(?P<width>\d+), deviceHeight=(?P<height>\d+).*')
        for line in self.shell('dumpsys display').splitlines():
            m = logicalDisplayRE.search(line, 0)
            if m:
                self.__displayInfo = {}
                for prop in [ 'width', 'height', 'orientation' ]:
                    self.__displayInfo[prop] = int(m.group(prop))
                for prop in [ 'density' ]:
                    d = self.__getDisplayDensity(None, strip=True, invokeGetPhysicalDisplayIfNotFound=True)
                    if d:
                        self.__displayInfo[prop] = d
                    else:
                        # No available density information
                        self.__displayInfo[prop] = -1.0
                return self.__displayInfo
        return None

    def getPhysicalDisplayInfo(self):
        ''' Gets C{mPhysicalDisplayInfo} values from dumpsys. This is a method to obtain display dimensions and density'''

        self.__checkTransport()
        phyDispRE = re.compile('Physical size: (?P<width>)x(?P<height>).*Physical density: (?P<density>)', re.MULTILINE)
        m = phyDispRE.search(self.shell('wm size; wm density'))
        if m:
            displayInfo = {}
            for prop in [ 'width', 'height' ]:
                displayInfo[prop] = int(m.group(prop))
            for prop in [ 'density' ]:
                displayInfo[prop] = float(m.group(prop))
            return displayInfo

        phyDispRE = re.compile('.*PhysicalDisplayInfo{(?P<width>\d+) x (?P<height>\d+), .*, density (?P<density>[\d.]+).*')
        for line in self.shell('dumpsys display').splitlines():
            m = phyDispRE.search(line, 0)
            if m:
                displayInfo = {}
                for prop in [ 'width', 'height' ]:
                    displayInfo[prop] = int(m.group(prop))
                for prop in [ 'density' ]:
                    # In mPhysicalDisplayInfo density is already a factor, no need to calculate
                    displayInfo[prop] = float(m.group(prop))
                return displayInfo

        # This could also be mSystem or mOverscanScreen
        phyDispRE = re.compile('\s*mUnrestrictedScreen=\((?P<x>\d+),(?P<y>\d+)\) (?P<width>\d+)x(?P<height>\d+)')
        # This is known to work on older versions (i.e. API 10) where mrestrictedScreen is not available
        dispWHRE = re.compile('\s*DisplayWidth=(?P<width>\d+) *DisplayHeight=(?P<height>\d+)')
        for line in self.shell('dumpsys window').splitlines():
            m = phyDispRE.search(line, 0)
            if not m:
                m = dispWHRE.search(line, 0)
            if m:
                displayInfo = {}
                for prop in [ 'width', 'height' ]:
                    displayInfo[prop] = int(m.group(prop))
                for prop in [ 'density' ]:
                    d = self.__getDisplayDensity(None, strip=True, invokeGetPhysicalDisplayIfNotFound=False)
                    if d:
                        displayInfo[prop] = d
                    else:
                        # No available density information
                        displayInfo[prop] = -1.0
                return displayInfo


    def __getProp(self, key, strip=True):
        if DEBUG:
            print >> sys.stderr, "__getProp(%s, %s)" % (key, strip)
        prop = self.shell('getprop %s' % key)
        if strip:
            prop = prop.rstrip('\r\n')
        if DEBUG:
            print >> sys.stderr, "    __getProp: returning '%s'" % prop
        return prop

    def __getDisplayWidth(self, key, strip=True):
        if self.__displayInfo and 'width' in self.__displayInfo:
            return self.__displayInfo['width']
        return self.getDisplayInfo()['width']

    def __getDisplayHeight(self, key, strip=True):
        if self.__displayInfo and 'height' in self.__displayInfo:
            return self.__displayInfo['height']
        return self.getDisplayInfo()['height']

    def __getDisplayOrientation(self, key, strip=True):
        if self.__displayInfo and 'orientation' in self.__displayInfo:
            return self.__displayInfo['orientation']
        displayInfo = self.getDisplayInfo()
        if 'orientation' in displayInfo:
            return displayInfo['orientation']
        # Fallback method to obtain the orientation
        # See https://github.com/dtmilano/AndroidViewClient/issues/128
        surfaceOrientationRE = re.compile('SurfaceOrientation:\s+(\d+)')
        output = self.shell('dumpsys input')
        m = surfaceOrientationRE.search(output)
        if m:
            return int(m.group(1))
        # We couldn't obtain the orientation
        return -1

    def __getDisplayDensity(self, key, strip=True, invokeGetPhysicalDisplayIfNotFound=True):
        if self.__displayInfo and 'density' in self.__displayInfo: # and self.__displayInfo['density'] != -1: # FIXME: need more testing
            return self.__displayInfo['density']
        BASE_DPI = 160.0
        d = self.getProperty('ro.sf.lcd_density', strip)
        if d:
            return float(d)/BASE_DPI
        d = self.getProperty('qemu.sf.lcd_density', strip)
        if d:
            return float(d)/BASE_DPI
        if invokeGetPhysicalDisplayIfNotFound:
            return self.getPhysicalDisplayInfo()['density']
        return -1.0

    def getSystemProperty(self, key, strip=True):
        self.__checkTransport()
        return self.getProperty(key, strip)

    def getProperty(self, key, strip=True):
        ''' Gets the property value for key '''

        self.__checkTransport()
        import collections
        MAP_PROPS = collections.OrderedDict([
                          (re.compile('display.width'), self.__getDisplayWidth),
                          (re.compile('display.height'), self.__getDisplayHeight),
                          (re.compile('display.density'), self.__getDisplayDensity),
                          (re.compile('display.orientation'), self.__getDisplayOrientation),
                          (re.compile('.*'), self.__getProp),
                          ])
        '''Maps properties key values (as regexps) to instance methods to obtain its values.'''

        for kre in MAP_PROPS.keys():
            if kre.match(key):
                return MAP_PROPS[kre](key=key, strip=strip)
        raise ValueError("key='%s' does not match any map entry")

    def getSdkVersion(self):
        '''
        Gets the SDK version.
        '''

        self.__checkTransport()
        return self.build[VERSION_SDK_PROPERTY]

    def press(self, name, eventType=DOWN_AND_UP):
        self.__checkTransport()
        if isinstance(name, unicode):
            name = name.decode('ascii', errors='replace')
        cmd = 'input keyevent %s' % name
        if DEBUG:
            print >> sys.stderr, "press(%s)" % cmd
        self.shell(cmd)

    def longPress(self, name, duration=0.5, dev='/dev/input/event0'):
        self.__checkTransport()
        # WORKAROUND:
        # Using 'input keyevent --longpress POWER' does not work correctly in
        # KitKat (API 19), it sends a short instead of a long press.
        # This uses the events instead, but it may vary from device to device.
        # The events sent are device dependent and may not work on other devices.
        # If this does not work on your device please do:
        #     $ adb shell getevent -l
        # and post the output to https://github.com/dtmilano/AndroidViewClient/issues
        # specifying the device and API level.
        if name[0:4] == 'KEY_':
            name = name[4:]
        if name in KEY_MAP:
            self.shell('sendevent %s 1 %d 1' % (dev, KEY_MAP[name]))
            self.shell('sendevent %s 0 0 0' % dev)
            time.sleep(duration)
            self.shell('sendevent %s 1 %d 0' % (dev, KEY_MAP[name]))
            self.shell('sendevent %s 0 0 0' % dev)
            return

        version = self.getSdkVersion()
        if version >= 19:
            cmd = 'input keyevent --longpress %s' % name
            if DEBUG:
                print >> sys.stderr, "longPress(%s)" % cmd
            self.shell(cmd)
        else:
            raise RuntimeError("longpress: not supported for API < 19 (version=%d)" % version)

    def startActivity(self, component=None, flags=None, uri=None):
        self.__checkTransport()
        cmd = 'am start'
        if component:
            cmd += ' -n %s' % component
        if flags:
            cmd += ' -f %s' % flags
        if uri:
            cmd += ' %s' % uri
        if DEBUG:
            print >> sys.stderr, "Starting activity: %s" % cmd
        out = self.shell(cmd)
        if re.search(r"(Error type)|(Error: )|(Cannot find 'App')", out, re.IGNORECASE | re.MULTILINE):
            raise RuntimeError(out)

    def takeSnapshot(self, reconnect=False):
        '''
        Takes a snapshot of the device and return it as a PIL Image.
        '''

        self.__checkTransport()
        try:
            from PIL import Image
        except:
            raise Exception("You have to install PIL to use takeSnapshot()")
        self.__send('framebuffer:', checkok=True, reconnect=False)
        import struct
        # case 1: // version
        #           return 12; // bpp, size, width, height, 4*(length, offset)
        received = self.__receive(1 * 4 + 12 * 4)
        (version, bpp, size, width, height, roffset, rlen, boffset, blen, goffset, glen, aoffset, alen) = struct.unpack('<' + 'L' * 13, received)
        if DEBUG:
            print >> sys.stderr, "    takeSnapshot:", (version, bpp, size, width, height, roffset, rlen, boffset, blen, goffset, glen, aoffset, alen)
        offsets = {roffset:'R', goffset:'G', boffset:'B'}
        if bpp == 32:
            if alen != 0:
                offsets[aoffset] = 'A'
            else:
                warnings.warn('''framebuffer is specified as 32bpp but alpha length is 0''')
        argMode = ''.join([offsets[o] for o in sorted(offsets)])
        if DEBUG:
            print >> sys.stderr, "    takeSnapshot:", (version, bpp, size, width, height, roffset, rlen, boffset, blen, goffset, blen, aoffset, alen, argMode)
        if argMode == 'BGRA':
            argMode = 'RGBA'
        if bpp == 16:
            mode = 'RGB'
            argMode += ';16'
        else:
            mode = argMode
        self.__send('\0', checkok=False, reconnect=False)
        if DEBUG:
            print >> sys.stderr, "    takeSnapshot: reading %d bytes" % (size)
        received = self.__receive(size)
        if reconnect:
            self.__connect()
            self.__setTransport()
        if DEBUG:
            print >> sys.stderr, "    takeSnapshot: Image.frombuffer(%s, %s, %s, %s, %s, %s, %s)" % (mode, (width, height), 'data', 'raw', argMode, 0, 1)
        image = Image.frombuffer(mode, (width, height), received, 'raw', argMode, 0, 1)
        # Just in case let's get the real image size
        if 'orientation' in self.display:
            r = (0, 90, 180, -90)[self.display['orientation']]
            image = image.rotate(r)
        return image

    def __transformPointByOrientation(self, (x, y), orientationOrig, orientationDest):
        if orientationOrig != orientationDest:
            if orientationDest == 1:
                _x = x
                x = self.display['width'] - y
                y = _x
            elif orientationDest == 3:
                _x = x
                x = y
                y = self.display['height'] - _x
        return (x, y)

    def touch(self, x, y, orientation=-1, eventType=DOWN_AND_UP):
        if DEBUG_TOUCH:
            print >> sys.stderr, "touch(x=", x, ", y=", y, ", orientation=", orientation, ", eventType=", eventType, ")"
        self.__checkTransport()
        if orientation == -1:
            orientation = self.display['orientation']
        self.shell('input tap %d %d' % self.__transformPointByOrientation((x, y), orientation, self.display['orientation']))

    def touchDip(self, x, y, orientation=-1, eventType=DOWN_AND_UP):
        if DEBUG_TOUCH:
            print >> sys.stderr, "touchDip(x=", x, ", y=", y, ", orientation=", orientation, ", eventType=", eventType, ")"
        self.__checkTransport()
        if orientation == -1:
            orientation = self.display['orientation']
        x = x * self.display['density']
        y = y * self.display['density']
        self.touch(x, y, orientation, eventType)

    def longTouch(self, x, y, duration=2000):
        '''
        Long touches at (x, y)
        
        @param duration: duration in ms
        
        This workaround was suggested by U{HaMi<http://stackoverflow.com/users/2571957/hami>}
        '''
        
        self.__checkTransport()
        self.drag((x, y), (x, y), duration, 1)

    def drag(self, (x0, y0), (x1, y1), duration, steps=1, orientation=-1):
        '''
        Sends drag event n PX (actually it's using C{input swipe} command.

        @param (x0, y0): starting point in PX
        @param (x1, y1): ending point in PX
        @param duration: duration of the event in ms
        @param steps: number of steps (currently ignored by @{input swipe}
        '''

        self.__checkTransport()
        if orientation == -1:
            orientation = self.display['orientation']
        (x0, y0) = self.__transformPointByOrientation((x0, y0), orientation, self.display['orientation'])
        (x1, y1) = self.__transformPointByOrientation((x1, y1), orientation, self.display['orientation'])

        version = self.getSdkVersion()
        if version <= 15:
            raise RuntimeError('drag: API <= 15 not supported (version=%d)' % version)
        elif version <= 17:
            self.shell('input swipe %d %d %d %d' % (x0, y0, x1, y1))
        else:
            self.shell('input touchscreen swipe %d %d %d %d %d' % (x0, y0, x1, y1, duration))

    def dragDip(self, (x0, y0), (x1, y1), duration, steps=1, orientation=-1):
        '''
        Sends drag event in DIP (actually it's using C{input swipe} command.

        @param (x0, y0): starting point in DIP
        @param (x1, y1): ending point in DIP
        @param duration: duration of the event in ms
        @param steps: number of steps (currently ignored by @{input swipe}
        '''

        self.__checkTransport()
        if orientation == -1:
            orientation = self.display['orientation']
        density = self.display['density'] if self.display['density'] > 0 else 1
        x0 = x0 * density
        y0 = y0 * density
        x1 = x1 * density
        y1 = y1 * density
        self.drag((x0, y0), (x1, y1), duration, steps, orientation)
        
    def type(self, text):
        self.__checkTransport()
        self.shell(u'input text "%s"' % text)

    def wake(self):
        self.__checkTransport()
        if not self.isScreenOn():
            self.shell('input keyevent POWER')

    def isLocked(self):
        '''
        Checks if the device screen is locked.

        @return True if the device screen is locked
        '''

        self.__checkTransport()
        lockScreenRE = re.compile('mShowingLockscreen=(true|false)')
        m = lockScreenRE.search(self.shell('dumpsys window policy'))
        if m:
            return (m.group(1) == 'true')
        raise RuntimeError("Couldn't determine screen lock state")

    def isScreenOn(self):
        '''
        Checks if the screen is ON.

        @return True if the device screen is ON
        '''

        self.__checkTransport()
        screenOnRE = re.compile('mScreenOnFully=(true|false)')
        m = screenOnRE.search(self.shell('dumpsys window policy'))
        if m:
            return (m.group(1) == 'true')
        raise RuntimeError("Couldn't determine screen ON state")

    def unlock(self):
        '''
        Unlocks the screen of the device.
        '''

        self.__checkTransport()
        self.shell('input keyevent MENU')
        self.shell('input keyevent BACK')

    @staticmethod
    def percentSame(image1, image2):
        '''
        Returns the percent of pixels that are equal

        @author: catshoes
        '''

        # If the images differ in size, return 0% same.
        size_x1, size_y1 = image1.size
        size_x2, size_y2 = image2.size
        if (size_x1 != size_x2 or
            size_y1 != size_y2):
            return 0

        # Images are the same size
        # Return the percent of pixels that are equal.
        numPixelsSame = 0
        numPixelsTotal = size_x1 * size_y1
        image1Pixels = image1.load()
        image2Pixels = image2.load()

        # Loop over all pixels, comparing pixel in image1 to image2
        for x in range(size_x1):
            for y in range(size_y1):
                if (image1Pixels[x, y] == image2Pixels[x, y]):
                    numPixelsSame += 1

        return numPixelsSame / float(numPixelsTotal)

    @staticmethod
    def sameAs(image1, image2, percent=1.0):
        '''
        Compares 2 images

        @author: catshoes
        '''

        return (AdbClient.percentSame(image1, image2) >= percent)

    def isKeyboardShown(self):
        '''
        Whether the keyboard is displayed.
        '''

        self.__checkTransport()
        dim = self.shell('dumpsys input_method')
        if dim:
            # FIXME: API >= 15 ?
            return "mInputShown=true" in dim
        return False

    def initDisplayProperties(self):
        self.__checkTransport()
        self.__displayInfo = None
        self.display['width'] = self.getProperty('display.width')
        self.display['height'] = self.getProperty('display.height')
        self.display['density'] = self.getProperty('display.density')
        self.display['orientation'] = self.getProperty('display.orientation')
        
    def log(self, tag, message, priority='D', verbose=False):
        if DEBUG_LOG:
            print >> sys.stderr, "log(tag=%s, message=%s, priority=%s, verbose=%s)" % (tag, message, priority, verbose)
        self.__checkTransport()
        message = self.substituteDeviceTemplate(message)
        if verbose or priority == 'V':
            print >> sys.stderr, tag+':', message
        self.shell('log -p %c -t "%s" %s' % (priority, tag, message))
        
    class __Log():
        '''
        Log class to simulate C{android.util.Log}
        '''

        def __init__(self, adbClient):
            self.adbClient = adbClient
            
        def __getattr__(self, attr):
            '''
            Returns the corresponding log method or @C{AttributeError}.
            '''
            
            if attr in ['v', 'd', 'i', 'w', 'e']:
                return lambda tag, message, verbose: self.adbClient.log(tag, message, priority=attr.upper(), verbose=verbose)
            raise AttributeError(self.__class__.__name__ + ' has no attribute "%s"' % attr)
    
    def getSystemService(self, name):
        if name == WIFI_SERVICE:
            return WifiManager(self)

    # def getWindows(self):
    #     self.__checkTransport()
    #     windows = {}
    #     dww = self.shell('dumpsys window windows')
    #     if DEBUG_WINDOWS: print >> sys.stderr, dww
    #     lines = dww.splitlines()
    #     widRE = re.compile('^ *Window #%s Window{%s (u\d+ )?%s?.*}:' %
    #                         (_nd('num'), _nh('winId'), _ns('activity', greedy=True)))
    #     currentFocusRE = re.compile('^  mCurrentFocus=Window{%s .*' % _nh('winId'))
    #     viewVisibilityRE = re.compile(' mViewVisibility=0x%s ' % _nh('visibility'))
    #     # This is for 4.0.4 API-15
    #     containingFrameRE = re.compile('^   *mContainingFrame=\[%s,%s\]\[%s,%s\] mParentFrame=\[%s,%s\]\[%s,%s\]' %
    #                          (_nd('cx'), _nd('cy'), _nd('cw'), _nd('ch'), _nd('px'), _nd('py'), _nd('pw'), _nd('ph')))
    #     contentFrameRE = re.compile('^   *mContentFrame=\[%s,%s\]\[%s,%s\] mVisibleFrame=\[%s,%s\]\[%s,%s\]' %
    #                          (_nd('x'), _nd('y'), _nd('w'), _nd('h'), _nd('vx'), _nd('vy'), _nd('vx1'), _nd('vy1')))
    #     # This is for 4.1 API-16
    #     framesRE = re.compile('^   *Frames: containing=\[%s,%s\]\[%s,%s\] parent=\[%s,%s\]\[%s,%s\]' %
    #                            (_nd('cx'), _nd('cy'), _nd('cw'), _nd('ch'), _nd('px'), _nd('py'), _nd('pw'), _nd('ph')))
    #     contentRE = re.compile('^     *content=\[%s,%s\]\[%s,%s\] visible=\[%s,%s\]\[%s,%s\]' %
    #                            (_nd('x'), _nd('y'), _nd('w'), _nd('h'), _nd('vx'), _nd('vy'), _nd('vx1'), _nd('vy1')))
    #     policyVisibilityRE = re.compile('mPolicyVisibility=%s ' % _ns('policyVisibility', greedy=True))

    #     for l in range(len(lines)):
    #         m = widRE.search(lines[l])
    #         if m:
    #             num = int(m.group('num'))
    #             winId = m.group('winId')
    #             activity = m.group('activity')
    #             wvx = 0
    #             wvy = 0
    #             wvw = 0
    #             wvh = 0
    #             px = 0
    #             py = 0
    #             visibility = -1
    #             policyVisibility = 0x0

    #             for l2 in range(l+1, len(lines)):
    #                 m = widRE.search(lines[l2])
    #                 if m:
    #                     l += (l2-1)
    #                     break
    #                 m = viewVisibilityRE.search(lines[l2])
    #                 if m:
    #                     visibility = int(m.group('visibility'))
    #                     if DEBUG_COORDS: print >> sys.stderr, "getWindows: visibility=", visibility
    #                 if self.build[VERSION_SDK_PROPERTY] >= 17:
    #                     wvx, wvy = (0, 0)
    #                     wvw, wvh = (0, 0)
    #                 if self.build[VERSION_SDK_PROPERTY] >= 16:
    #                     m = framesRE.search(lines[l2])
    #                     if m:
    #                         px, py = obtainPxPy(m)
    #                         m = contentRE.search(lines[l2+1])
    #                         if m:
    #                             # FIXME: the information provided by 'dumpsys window windows' in 4.2.1 (API 16)
    #                             # when there's a system dialog may not be correct and causes the View coordinates
    #                             # be offset by this amount, see
    #                             # https://github.com/dtmilano/AndroidViewClient/issues/29
    #                             wvx, wvy = obtainVxVy(m)
    #                             wvw, wvh = obtainVwVh(m)
    #                 elif self.build[VERSION_SDK_PROPERTY] == 15:
    #                     m = containingFrameRE.search(lines[l2])
    #                     if m:
    #                         px, py = obtainPxPy(m)
    #                         m = contentFrameRE.search(lines[l2+1])
    #                         if m:
    #                             wvx, wvy = obtainVxVy(m)
    #                             wvw, wvh = obtainVwVh(m)
    #                 elif self.build[VERSION_SDK_PROPERTY] == 10:
    #                     m = containingFrameRE.search(lines[l2])
    #                     if m:
    #                         px, py = obtainPxPy(m)
    #                         m = contentFrameRE.search(lines[l2+1])
    #                         if m:
    #                             wvx, wvy = obtainVxVy(m)
    #                             wvw, wvh = obtainVwVh(m)
    #                 else:
    #                     warnings.warn("Unsupported Android version %d" % self.build[VERSION_SDK_PROPERTY])

    #                 #print >> sys.stderr, "Searching policyVisibility in", lines[l2]
    #                 m = policyVisibilityRE.search(lines[l2])
    #                 if m:
    #                     policyVisibility = 0x0 if m.group('policyVisibility') == 'true' else 0x8

    #             windows[winId] = Window(num, winId, activity, wvx, wvy, wvw, wvh, px, py, visibility + policyVisibility)
    #         else:
    #             m = currentFocusRE.search(lines[l])
    #             if m:
    #                 currentFocus = m.group('winId')
        
    #     if currentFocus in windows and windows[currentFocus].visibility == 0:
    #         if DEBUG_COORDS:
    #             print >> sys.stderr, "getWindows: focus=", currentFocus
    #             print >> sys.stderr, "getWindows:", windows[currentFocus]
    #         windows[currentFocus].focused = True

    #     return windows
    
    # def getFocusedWindow(self):
    #     '''
    #     Gets the focused window.
        
    #     @return: The focused L{Window}.
    #     '''

    #     for window in self.getWindows().values():
    #         if window.focused:
    #             return window
    #     return None
    
    # def getFocusedWindowName(self):
    #     '''
    #     Gets the focused window name.
        
    #     This is much like monkeyRunner's C{HierarchyView.getWindowName()}
        
    #     @return: The focused window name
    #     '''

    #     window = self.getFocusedWindow()
    #     if window:
    #         return window.activity
    #     return None

    # def substituteDeviceTemplate(self, template):
    #     serialno = self.serialno.replace('.', '_').replace(':', '-')
    #     focusedWindowName = self.getFocusedWindowName().replace('/', '-').replace('.', '_')
    #     timestamp = datetime.datetime.now().isoformat()
    #     _map = {
    #             'serialno': serialno,
    #             'focusedwindowname': focusedWindowName,
    #             'timestamp': timestamp
    #          }
    #     return string.Template(template).substitute(_map)

if __name__ == '__main__':
    adbClient = AdbClient(os.environ['ANDROID_SERIAL'])
    INTERACTIVE = False
    if INTERACTIVE:
        sout = adbClient.shell()
        prompt = re.compile(".+@android:(.*) [$#] \r\r\n")
        while True:
            try:
                cmd = raw_input('adb $ ')
            except EOFError:
                break
            if cmd == 'exit':
                break
            adbClient.socket.__send(cmd + "\r\n")
            sout.readline(4096)  # eat first line, which is the command
            while True:
                line = sout.readline(4096)
                if prompt.match(line):
                    break
                print line,
                if not line:
                    break

        print "\nBye"
    else:
        print 'date:', adbClient.shell('date')



#!/usr/bin/env python

# USAGE
# python click_and_crop.py --image jurassic_park_kitchen.jpg

# import the necessary packages
import os
import sys
import argparse
import threading
import Tkinter
import tkSimpleDialog
import collections
from Queue import Queue
from StringIO import StringIO
import cv2
from PIL import ImageTk, Image

import atx


Point = collections.namedtuple('Point', ['x', 'y'])

def cv2_to_pil(image):
    img_str = cv2.imencode('.png', image)[1].tostring()
    return Image.open(StringIO(img_str))


def make_mouse_callback(imgs, ref_pt):
    # initialize the list of reference points and boolean indicating
    # whether cropping is being performed or not
    cropping = [False]

    def _click_and_crop(event, x, y, flags, param):
        # grab references to the global variables
        # global ref_pt, cropping

        # if the left mouse button was clicked, record the starting
        # (x, y) coordinates and indicate that cropping is being
        # performed
        if event == cv2.EVENT_LBUTTONDOWN:
            ref_pt[0] = Point(x, y)
            cropping[0] = True

        # check to see if the left mouse button was released
        elif event == cv2.EVENT_LBUTTONUP:
            # record the ending (x, y) coordinates and indicate that
            # the cropping operation is finished
            ref_pt[1] = Point(x, y)
            cropping[0] = False

            # draw a rectangle around the region of interest
            imgs[1] = image = imgs[0].copy()
            cv2.rectangle(image, ref_pt[0], ref_pt[1], (0, 255, 0), 2)
            cv2.imshow("image", image)
        elif event == cv2.EVENT_MOUSEMOVE and cropping[0]:
            img2 = imgs[0].copy()
            cv2.rectangle(img2, ref_pt[0], (x, y), (0, 255, 0), 2)
            imgs[1] = image = img2
            cv2.imshow("image", image)
    return _click_and_crop

def interactive_save(image): #, save_to=None):
    imgpil = cv2_to_pil(image)
    #img_str = cv2.imencode('.png', image)[1].tostring()
    #imgpil = Image.open(StringIO(img_str))

    root = Tkinter.Tk()
    root.geometry('{}x{}'.format(400, 400))
    imgtk = ImageTk.PhotoImage(image=imgpil)
    panel = Tkinter.Label(root, image=imgtk) #.pack()
    panel.pack(side="bottom", fill="both", expand="yes")
    
    # if not save_to:
    save_to = tkSimpleDialog.askstring("Save cropped image", "Enter filename")
    if save_to:
        if save_to.find('.') == -1:
            save_to += '.png'
        print 'Save to:', save_to
        cv2.imwrite(save_to, image)
    root.destroy()

def long_task(secs=2):
    import time
    time.sleep(secs)
    print 'hello'

def worker(que):
    while True:
        (func, args, kwargs) = que.get()
        try:
            func(*args, **kwargs)
        except Exception as e:
            print e
        finally:
            que.task_done()

def simple_ide():
    que = Queue()
    que.put((long_task, (), {}))

    th = threading.Thread(name='worker', target=worker, args=(que,))
    th.daemon = True
    th.start()

    root = Tkinter.Tk()
    root.geometry('{}x{}'.format(400, 400))
    btn_hello = Tkinter.Button(root, text="Hello!")
    #btn_hello.set_text("World")
    btn_hello.pack()
    root.mainloop()

def main():
    # construct the argument parser and parse the arguments
    #simple_ide()
    ap = argparse.ArgumentParser()
    ap.add_argument("-s", "--serial", required=False, help="Android serialno")
    # ap.add_argument("-o", "--output", required=True, help="Output image file, ext must be png")
    args = vars(ap.parse_args())

    d = atx.connect(args["serial"])
    origin = d.screenshot()

    # load the image, clone it, and setup the mouse callback function
    # origin = cv2.imread(output)
    image = cv2.resize(origin, (0, 0), fx=0.5, fy=0.5) 
    images = [image.copy(), image]
    ref_pt = [None, None]
    
    cv2.namedWindow("image")
    cv2.setMouseCallback("image", make_mouse_callback(images, ref_pt))

    while True:
        # display the image and wait for a keypress
        cv2.imshow("image", images[1])
        key = cv2.waitKey(1) & 0xFF

        # if the 'c' key is pressed, break from the loop
        if key == ord("r"):
            origin = d.screenshot()
            image = cv2.resize(origin, (0, 0), fx=0.5, fy=0.5) 
            images[0] = image.copy()
            images[1] = image
        if key == ord("c"):
            if ref_pt[1] is None:
                continue
            (a, b) = ref_pt
            a = Point(a.x*2, a.y*2)
            b = Point(b.x*2, b.y*2)
            roi = origin[a.y:b.y, a.x:b.x]
            interactive_save(roi)
        elif key == ord("q"):
            break

    # close all open windows
    cv2.destroyAllWindows()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import flask

import airtest
import airtest.image as aim
import cv2
import time

from . import utils

app = None #airtest.connect(monitor=False)

bp = flask.Blueprint('api', __name__)

@bp.route('/')
def home():
    return 'API documentation'

@bp.route('/snapshot')
def snapshot():
    filename = '%d-screen.png' % int(time.time())
    if os.path.exists(filename):
        os.unlink(filename)
    app.takeSnapshot(os.path.join(utils.TMPDIR, filename))
    return flask.jsonify(dict(filename=filename))

@bp.route('/crop')
def crop():
    rget = flask.request.args.get

    filename = rget('filename')
    screen = rget('screen')
    x, y = int(rget('x')), int(rget('y'))
    width, height = int(rget('width')), int(rget('height'))

    screen_file = screen.lstrip('/').replace('/', os.sep)
    screen_path = os.path.join(utils.selfdir(), screen_file)
    output_path = os.path.join(utils.workdir(), filename)
    assert os.path.exists(screen_path)

    im = cv2.imread(screen_path)
    cv2.imwrite(output_path, im[y:y+height, x:x+width])
    return flask.jsonify(dict(success=True, 
        message="文件已保存: "+output_path.encode('utf-8')))

@bp.route('/cropcheck')
def crop_check():
    rget = flask.request.args.get
    
    screen = rget('screen')
    x, y = int(rget('x')), int(rget('y'))
    width, height = int(rget('width')), int(rget('height'))

    screen_file = screen.lstrip('/').replace('/', os.sep)
    screen_path = os.path.join(utils.selfdir(), screen_file)

    im = cv2.imread(screen_path)
    im = im[y:y+height, x:x+width]  # crop image
    siftcnt = aim.sift_point_count(im)
    return flask.jsonify(dict(siftcnt=siftcnt))

@bp.route('/run')
def run_code():
    global app
    code = flask.request.args.get('code')
    try:
        exec code        
    except Exception, e:
        return flask.jsonify(dict(success=False, message=str(e)))
    return flask.jsonify(dict(success=True, message=""))

@bp.route('/connect')
def connect():
    global app
    device = flask.request.args.get('device')
    devno = flask.request.args.get('devno')
    try:
        app = airtest.connect(devno, device=device, monitor=False)
    except Exception, e:
        return flask.jsonify(dict(success=False, message=str(e)))
        
    return flask.jsonify(dict(success=True, message="连接成功"))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#coding: utf-8

import sys
reload(sys)
sys.setdefaultencoding('utf8')

import airtest
import os
import cv2
import time

app = airtest.connect('shit', device='dummy')
screenFile = 'testdata/dummy/default.png'
app.globalSet(image_dirs=['testdata/dummy', '.'])
app.globalSet(image_match_method='template')
app.globalSet(threshold=0.8)

ICON = 'testdata/dummy/add.png'

def setup_function(f):
    app.dev._snapshot = screenFile
    app.dev._text = ''
    app.dev._click = None

def test_locate_image():
    # ap = airtest.connect('test-connect', appname='hello', device='dummy', monitor=False, logfile='log/1')
    app.dev._snapshot = 'testdata/dummy/h15_bg.png'
    print 'get method=', app.globalGet('image_match_method')
    print app.click('back.png')

def test_connect_monitor():
    ap = airtest.connect('test-connect', appname='hello', device='dummy', monitor=False, logfile='log/1')
    ap.dev._getCpu = False
    time.sleep(2.0)
    assert ap.dev._getCpu == False
    ap.globalSet(enable_monitor=False)

    ap = airtest.connect('test-connect', appname='hello', device='dummy', monitor=True, logfile='log/2')
    ap.dev._getCpu = False
    time.sleep(2.0)
    assert ap.dev._getCpu == True
    ap.globalSet(enable_monitor=False)

    ap = airtest.connect('test-connect', appname='hello', device='dummy', logfile='log/3')
    ap.dev._getCpu = False
    time.sleep(2.0)
    assert ap.dev._getCpu == True
    ap.globalSet(enable_monitor=False)
    
def test_snapshot():
    app.takeSnapshot('tmp/nice.png')
    assert os.path.exists('tmp/nice.png')
    os.unlink('tmp/nice.png')

def test_find():
    pos = app.find('aboy')
    assert pos == None

def test_type():
    app.type('hello')
    assert app.dev._text == 'hello'

def test_shape():
    x, y = app.shape()
    assert isinstance(x, int)
    assert isinstance(y, int)

def test_click():
    app.click(u'赵云')
    x, y = app.dev._click
    assert 200 < x < 250
    assert 400 < y < 450

    app.click(u'赵云.png')
    x, y = app.dev._click
    assert 200 < x < 250
    assert 400 < y < 450

    app.click((10, 10))
    assert app.dev._click == (10, 10)

    app.click((0.4, 0.5))
    w, h = app.shape()
    assert app.dev._click == (int(0.4*w), int(0.5*h))


def other():
    scr = cv2.imread(screenFile, 0)
    icon = cv2.imread(iconFile, 0)

    res = cv2.matchTemplate(icon, scr, cv2.TM_CCOEFF)
    minVal, maxVal, minLoc, maxLoc = cv2.minMaxLoc(res)
    topLeft = maxLoc
    print topLeft



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module helps make mobile test easily
"""

import sys
import signal

import patch
import device
import pkg_resources
try:
    version = pkg_resources.get_distribution("atx").version
except pkg_resources.DistributionNotFound:
    version = 'unknown'

from atx.consts import *
from atx.errors import *
from atx.device import Watcher


def connect(serialno, platform='android'):
    """Connect to a device, and return its object
    Args:
        platform: string one of <android|ios|windows>

    Returns:
        None
    """
    devclss = {
        'android': device.AndroidDevice,
    }
    cls = devclss.get(platform)
    if cls is None:
        raise Exception('Platform not exists')
    return cls(serialno)


# import os
# import json
# import subprocess
# import signal
# import sys
# import os

# from . import patch
# from . import proto
# from . import cron

# def _sig_handler(signum, frame):
#     print >>sys.stderr, 'Signal INT catched !!!'
#     sys.exit(1)
# signal.signal(signal.SIGINT, _sig_handler)


# # defaultConfigFile = 'air.json'
# defaultDevice = 'android'


# def _parse_addr(addr):
#     '''
#     通过 proto://url 拿到相应的模块和地址
#     '''
#     import urlparse
#     p = urlparse.urlparse(addr)
#     if not p.scheme:
#         raise RuntimeError('device type must be specified')

#     # FIXME(ssx): better change to __import__
#     exec('from .device import '+p.scheme)#, p.netloc, p.path
#     module = eval(p.scheme)
#     # 自动查找设备
#     loc = p.netloc
#     if p.scheme == 'android' and not loc:
#         loc = mustOneDevice()
#     return module, loc, p


# class Device(object):
#     '''
#     Create a new device instance and use this instance to control device
#     '''
#     def __init__(self, addr, logfile=None):
#         '''
#         @addr: eg: android://<serialno> or ios://127.0.0.1
#         '''
#         from . import devsuit
#         module, loc, p = _parse_addr(addr)
#         if p.path and p.netloc:
#             serialno = p.path.lstrip('/')
#             addr = p.netloc
#         else:
#             serialno = p.netloc
#             addr = ''
#         dev = module.Device(serialno, addr) # FIXME(ssx): may not fit well with ios
#         self._m = devsuit.DeviceSuit(p.scheme, dev, logfile=logfile)

#     def __getattr__(self, key):
#         if hasattr(self._m, key):
#             return getattr(self._m, key)
#         raise AttributeError('Monitor object has no attribute "%s"' % key)

# class _JoinClass(object):
#     def __init__(self, clss):
#         ''' clss ~ list:classes '''
#         self._clss = clss

#     def __getattr__(self, key):
#         for cl in self._clss:
#             if hasattr(cl, key):
#                 return getattr(cl, key)
#         raise AttributeError('Object has no attribute "%s"' % key)

# def connect(addr='', appname=None, device=None, interval=3.0, logfile='log/atx.log'):
#     clss = []
#     # compatible with old connect style
#     if addr.find('://') == -1:
#         addr = (device or defaultDevice) +'://'+addr

#     dev = Device(addr, logfile)
#     clss.append(dev)

#     c = _JoinClass(clss)

#     @patch.attachmethod(c)
#     def logPerformance(self):
#         if not hasattr(m, 'cpu'):
#             return
#         dev.log(proto.TAG_CPU, m.cpu())
#         dev.log(proto.TAG_MEMORY, m.memory())
#         dev.log(proto.TAG_BATTERY, m.battery())
#         dev.log(proto.TAG_SYS_MEMORY, m.sys_memory())

#     @patch.attachmethod(c)
#     def startMonitor(self):
#         c.cron.start()

#     @patch.attachmethod(c)
#     def stopMonitor(self):
#         c.cron.stop()
#         print 'stop monitor'

#     c.cron = cron.Crontab()
#     c.cron.addfunc(c.logPerformance)
#     c.cron._cycle = interval

#     if monitor:
#         c.startMonitor()
    
#     return c

# def getDevices(device='android'):
#     ''' 
#     @return devices list 
#     '''
#     subprocess.call(['adb', 'start-server'])
#     output = subprocess.check_output(['adb', 'devices'])
#     result = []
#     for line in str(output).splitlines()[1:]:
#         ss = line.strip().split()
#         if len(ss) == 2:
#             (devno, state) = ss
#             result.append((devno, state))
#     return result

# def mustOneDevice():
#     ''' make sure only one devices connected '''
#     serialno = os.getenv(proto.ENV_SERIALNO)
#     if serialno:
#         return serialno

#     devs = [d for d, t in getDevices() if t == 'device']
#     if len(devs) == 0:
#         raise RuntimeError('no device connected')
#     if len(devs) > 1:
#         raise RuntimeError('must specify one device')
#     return devs[0]



import os

def selfdir():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.dirname(cwd)

def workdir():
    return os.getenv("WORKDIR")

TMPDIR = os.path.join(selfdir(), 'tmp')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
convert log to html report
'''

import fuckit

import os
import sys
import json
import time
import shutil

import pystache
import markupsafe

from airtest import base

TIME_FORMAT = '%Y/%m/%d %H:%M:%S'

def render(logfile, htmldir):
    '''
    parse logfile and render it to html
    '''
    if not os.path.exists(logfile):
        sys.exit('logfile: %s not exists' %(logfile))
    if not os.path.exists(htmldir):
        os.makedirs(htmldir)
    cpus, items = [], []
    fpss = []
    mems = []
    imgs = []
    # The render data
    data = {
            'info': {
                'generated_time': time.strftime('%Y/%m/%d %H:%M:%S'),
            },
            'items': items, 
            'cpu_data': None,
            'mem_data': None,
            'images': imgs,
        }
    info = data.get('info')

    # Read log line by line
    from . import proto
    records = []
    for line in open(logfile):
        v = json.loads(line)
        r = {'time': time.strftime(TIME_FORMAT, time.localtime(v.get('timestamp')))}
        d = v.get('data', {})
        tag = v.get('tag')

        # Process Function, Snapshot, Memory, CPU ...
        if tag == proto.TAG_FUNCTION:
            tag = markupsafe.Markup('function')    
            args = map(json.dumps, d.get('args'))
            kwargs = [ '%s=%s' %(k, json.dumps(_v)) for k, _v in d.get('kwargs', {}).items() ]
            message = '<code style="color:green">%s(%s)</code>' %(d.get('name'), ', '.join(args+kwargs))
            message = markupsafe.Markup(message)
        elif tag == proto.TAG_SNAPSHOT:
            message = markupsafe.Markup("<img width=100%% src='%s'/>" % d.get('filename'))
        elif tag == proto.TAG_CPU:
            message = '%d%%' %(d)
            cpus.append([r, d])
        elif tag == proto.TAG_MEMORY:
            mems.append([r, d['PSS']])
            message = json.dumps(d)
        else:
            message = None
        
        if message:
            r['tag'] = tag
            r['message'] = message
            records.append(r)

    # Calculate average cpu and mem
    data['records'] = records
    def average(ss):
        if ss:
            return reduce(lambda x,y: x+y, [value for _,value in ss])/float(len(ss))
        return 0.0

    data['cpu_average'] = round(average(cpus), 2)
    data['mem_average'] = round(average(mems), 2)
    data['fps_average'] = round(average(fpss), 2)

    tmpldir = os.path.join(base.dirname(__file__), 'htmltemplate')
    for name in os.listdir(tmpldir):
        fullpath = os.path.join(tmpldir, name)
        outpath = os.path.join(htmldir, name)
        if os.path.isdir(fullpath):
            shutil.rmtree(outpath, ignore_errors=True)
            shutil.copytree(fullpath, outpath)
            continue
        if fullpath.endswith('.swp'):
            continue
        content = open(fullpath).read().decode('utf-8')
        out = pystache.Renderer(escape=markupsafe.escape).render(content, data)
        print fullpath
        with open(outpath, 'w') as file:
            file.write(out.encode('utf-8'))

        # store json data file, for other system
        with open(os.path.join(htmldir, 'data.json'), 'w') as file:
            json.dump(data, file)
    # Copy snapshots
    if htmldir != '.':
        shutil.rmtree(os.path.join(htmldir, 'tmp'), ignore_errors=True)
    shutil.copytree('tmp', os.path.join(htmldir, 'tmp'))

if __name__ == '__main__':
    render('testdata/airtest.log', 'tmp/out.html')



# coding: utf-8

import os
import time
import atx
from atx import consts
print 'Version:', atx.version

d = atx.connect(None)

# print d.screenshot('screen.png')

def screenshot():
    # height = d.info['displayHeight']
    # height = 1920 # this is for my phone
    # params = '{x}x{y}@{x}x{y}/{r}'.format(
    #     x=d.info['displayWidth'], y=height, r=d.info['displayRotation'])
    # start = time.time()
    # command = 'LD_LIBRARY_PATH=/data/local/tmp /data/local/tmp/minicap -P {} -s > /data/local/tmp/screen2.png'.format(params)
    # # print d.touch_image('a.png')
    # d.shell(command)
    # os.system('adb pull /data/local/tmp/screen2.png')
    # print time.time()-start
    # Image now save in screen2.png
    # d._minicap()
    start = time.time()
    # d.screenshot_method = consts.SCREENSHOT_METHOD_MINICAP
    d.screenshot('ttt.png')
    print time.time() - start

def start_app():
    d.start_app('com.netease.txx')

def stop_app():
    d.stop_app('com.netease.txx', clear=True)

def touch():
    d.screenshot_method = consts.SCREENSHOT_METHOD_MINICAP
    d.touch_image('button.png')

if __name__ == '__main__':
    # stop_app()
    w = atx.Watcher()
    w.on('setting.png', atx.Watcher.ACTION_TOUCH)
    w.on('common.png', atx.Watcher.ACTION_TOUCH)

    wid = d.add_watcher(w)
    d.del_watcher(wid)
    while 1:
        screenshot()
    # screenshot()
    # touch()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# airtest web gui
#
import os
import flask
from .routers import utils

app = flask.Flask(__name__)

app.config['DEBUG'] = True


from .routers import home, api
app.register_blueprint(home.bp, url_prefix='')
app.register_blueprint(api.bp, url_prefix='/api')

def serve(*args, **kwargs):
    # create tmpdir if not exists
    if not os.path.exists(utils.TMPDIR):
        os.makedirs(utils.TMPDIR)
    print 'Clean tempfiles ...'
    for file in os.listdir(utils.TMPDIR):
        filepath = os.path.join(utils.TMPDIR, file)
        if os.path.isfile(filepath):
            os.unlink(filepath)
    app.run(*args, **kwargs)

if __name__ == '__main__':
    serve()


# coding: utf8

import flask
# from flask import request, flash, redirect, url_for, render_template

from . import utils

bp = flask.Blueprint('home', __name__)

@bp.route('/tmp/<path:path>')
def static_proxy(path):
    return flask.send_from_directory(utils.TMPDIR, path)

#@models.db_session
@bp.route('/')
def home():
    return flask.render_template('index.html')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# export PYTHONPATH=/z/workspace/airtest

import airtest
import pytest

def test_android_getsysinfo():
    devs = airtest.getDevices()
    if not devs:
        pytest.skip('not devices detected')
    if devs:
        phoneNo, phoneType = devs[0]
    print phoneNo
    app = airtest.connect(phoneNo, device='android')
    print app.dev.getdevinfo()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import platform

from airtest import base

def test_exec_cmd():
    if platform.system() == 'Linux':
        base.exec_cmd('echo', 'hello')

def test_exec_cmd_shell():
    base.exec_cmd('echo hello', shell=True)

def test_check_output():
    output = base.check_output('echo hello')
    assert output.rstrip() == 'hello'




'''
Created on Jun 13, 2014

@author: diego
'''

KEY_MAP = dict()
KEY_MAP["ESCAPE"]=1
KEY_MAP["1"]=2
KEY_MAP["2"]=3
KEY_MAP["3"]=4
KEY_MAP["4"]=5
KEY_MAP["5"]=6
KEY_MAP["6"]=7
KEY_MAP["7"]=8
KEY_MAP["8"]=9
KEY_MAP["9"]=10
KEY_MAP["0"]=11
KEY_MAP["MINUS"]=12
KEY_MAP["EQUALS"]=13
KEY_MAP["DEL"]=14
KEY_MAP["TAB"]=15
KEY_MAP["Q"]=16
KEY_MAP["W"]=17
KEY_MAP["E"]=18
KEY_MAP["R"]=19
KEY_MAP["T"]=20
KEY_MAP["Y"]=21
KEY_MAP["U"]=22
KEY_MAP["I"]=23
KEY_MAP["O"]=24
KEY_MAP["P"]=25
KEY_MAP["LEFT_BRACKET"]=26
KEY_MAP["RIGHT_BRACKET"]=27
KEY_MAP["ENTER"]=28
KEY_MAP["CTRL_LEFT"]=29
KEY_MAP["A"]=30
KEY_MAP["S"]=31
KEY_MAP["D"]=32
KEY_MAP["F"]=33
KEY_MAP["G"]=34
KEY_MAP["H"]=35
KEY_MAP["J"]=36
KEY_MAP["K"]=37
KEY_MAP["L"]=38
KEY_MAP["SEMICOLON"]=39
KEY_MAP["APOSTROPHE"]=40
KEY_MAP["GRAVE"]=41
KEY_MAP["SHIFT_LEFT"]=42
KEY_MAP["BACKSLASH"]=43
KEY_MAP["Z"]=44
KEY_MAP["X"]=45
KEY_MAP["C"]=46
KEY_MAP["V"]=47
KEY_MAP["B"]=48
KEY_MAP["N"]=49
KEY_MAP["M"]=50
KEY_MAP["COMMA"]=51
KEY_MAP["PERIOD"]=52
KEY_MAP["SLASH"]=53
KEY_MAP["SHIFT_RIGHT"]=54
KEY_MAP["NUMPAD_MULTIPLY"]=55
KEY_MAP["ALT_LEFT"]=56
KEY_MAP["SPACE"]=57
KEY_MAP["CAPS_LOCK"]=58
KEY_MAP["F1"]=59
KEY_MAP["F2"]=60
KEY_MAP["F3"]=61
KEY_MAP["F4"]=62
KEY_MAP["F5"]=63
KEY_MAP["F6"]=64
KEY_MAP["F7"]=65
KEY_MAP["F8"]=66
KEY_MAP["F9"]=67
KEY_MAP["F10"]=68
KEY_MAP["NUM_LOCK"]=69
KEY_MAP["SCROLL_LOCK"]=70
KEY_MAP["NUMPAD_7"]=71
KEY_MAP["NUMPAD_8"]=72
KEY_MAP["NUMPAD_9"]=73
KEY_MAP["NUMPAD_SUBTRACT"]=74
KEY_MAP["NUMPAD_4"]=75
KEY_MAP["NUMPAD_5"]=76
KEY_MAP["NUMPAD_6"]=77
KEY_MAP["NUMPAD_ADD"]=78
KEY_MAP["NUMPAD_1"]=79
KEY_MAP["NUMPAD_2"]=80
KEY_MAP["NUMPAD_3"]=81
KEY_MAP["NUMPAD_0"]=82
KEY_MAP["NUMPAD_DOT"]=83
KEY_MAP["ZENKAKU_HANKAKU"]=85
KEY_MAP["BACKSLASH"]=86
KEY_MAP["F11"]=87
KEY_MAP["F12"]=88
KEY_MAP["RO"]=89
KEY_MAP["HENKAN"]=92
KEY_MAP["KATAKANA_HIRAGANA"]=93
KEY_MAP["MUHENKAN"]=94
KEY_MAP["NUMPAD_COMMA"]=95
KEY_MAP["NUMPAD_ENTER"]=96
KEY_MAP["CTRL_RIGHT"]=97
KEY_MAP["NUMPAD_DIVIDE"]=98
KEY_MAP["SYSRQ"]=99
KEY_MAP["ALT_RIGHT"]=100
KEY_MAP["MOVE_HOME"]=102
KEY_MAP["DPAD_UP"]=103
KEY_MAP["PAGE_UP"]=104
KEY_MAP["DPAD_LEFT"]=105
KEY_MAP["DPAD_RIGHT"]=106
KEY_MAP["MOVE_END"]=107
KEY_MAP["DPAD_DOWN"]=108
KEY_MAP["PAGE_DOWN"]=109
KEY_MAP["INSERT"]=110
KEY_MAP["FORWARD_DEL"]=111
KEY_MAP["VOLUME_MUTE"]=113
KEY_MAP["VOLUME_DOWN"]=114
KEY_MAP["VOLUME_UP"]=115
KEY_MAP["POWER"]=116
KEY_MAP["NUMPAD_EQUALS"]=117
KEY_MAP["BREAK"]=119
KEY_MAP["NUMPAD_COMMA"]=121
KEY_MAP["KANA"]=122
KEY_MAP["EISU"]=123
KEY_MAP["YEN"]=124
KEY_MAP["META_LEFT"]=125
KEY_MAP["META_RIGHT"]=126
KEY_MAP["MENU"]=127
KEY_MAP["MEDIA_STOP"]=128
KEY_MAP["MENU"]=139
KEY_MAP["CALCULATOR"]=140
#KEY_MAP["POWER"]=142
#KEY_MAP["POWER"]=143
KEY_MAP["EXPLORER"]=150
#KEY_MAP["POWER"]=152
KEY_MAP["ENVELOPE"]=155
KEY_MAP["BOOKMARK"]=156
KEY_MAP["BACK"]=158
KEY_MAP["FORWARD"]=159
KEY_MAP["MEDIA_CLOSE"]=160
KEY_MAP["MEDIA_EJECT"]=161
KEY_MAP["MEDIA_EJECT"]=162
KEY_MAP["MEDIA_NEXT"]=163
KEY_MAP["MEDIA_PLAY_PAUSE"]=164
KEY_MAP["MEDIA_PREVIOUS"]=165
KEY_MAP["MEDIA_STOP"]=166
KEY_MAP["MEDIA_RECORD"]=167
KEY_MAP["MEDIA_REWIND"]=168
KEY_MAP["CALL"]=169
KEY_MAP["MUSIC"]=171
KEY_MAP["HOME"]=172
KEY_MAP["PAGE_UP"]=177
KEY_MAP["PAGE_DOWN"]=178
KEY_MAP["NUMPAD_LEFT_PAREN"]=179
KEY_MAP["NUMPAD_RIGHT_PAREN"]=180
KEY_MAP["MEDIA_PLAY"]=200
KEY_MAP["MEDIA_PAUSE"]=201
KEY_MAP["MEDIA_PLAY"]=207
KEY_MAP["MEDIA_FAST_FORWARD"]=208
KEY_MAP["CAMERA"]=212
KEY_MAP["MUSIC"]=213
KEY_MAP["ENVELOPE"]=215
KEY_MAP["SEARCH"]=217
KEY_MAP["BRIGHTNESS_DOWN"]=224
KEY_MAP["BRIGHTNESS_UP"]=225
KEY_MAP["HEADSETHOOK"]=226
KEY_MAP["BUTTON_1"]=256
KEY_MAP["BUTTON_2"]=257
KEY_MAP["BUTTON_3"]=258
KEY_MAP["BUTTON_4"]=259
KEY_MAP["BUTTON_5"]=260
KEY_MAP["BUTTON_6"]=261
KEY_MAP["BUTTON_7"]=262
KEY_MAP["BUTTON_8"]=263
KEY_MAP["BUTTON_9"]=264
KEY_MAP["BUTTON_10"]=265
KEY_MAP["BUTTON_11"]=266
KEY_MAP["BUTTON_12"]=267
KEY_MAP["BUTTON_13"]=268
KEY_MAP["BUTTON_14"]=269
KEY_MAP["BUTTON_15"]=270
KEY_MAP["BUTTON_16"]=271
KEY_MAP["BUTTON_1"]=288
KEY_MAP["BUTTON_2"]=289
KEY_MAP["BUTTON_3"]=290
KEY_MAP["BUTTON_4"]=291
KEY_MAP["BUTTON_5"]=292
KEY_MAP["BUTTON_6"]=293
KEY_MAP["BUTTON_7"]=294
KEY_MAP["BUTTON_8"]=295
KEY_MAP["BUTTON_9"]=296
KEY_MAP["BUTTON_10"]=297
KEY_MAP["BUTTON_11"]=298
KEY_MAP["BUTTON_12"]=299
KEY_MAP["BUTTON_13"]=300
KEY_MAP["BUTTON_14"]=301
KEY_MAP["BUTTON_15"]=302
KEY_MAP["BUTTON_16"]=303
KEY_MAP["BUTTON_A"]=304
KEY_MAP["BUTTON_B"]=305
KEY_MAP["BUTTON_C"]=306
KEY_MAP["BUTTON_X"]=307
KEY_MAP["BUTTON_Y"]=308
KEY_MAP["BUTTON_Z"]=309
KEY_MAP["BUTTON_L1"]=310
KEY_MAP["BUTTON_R1"]=311
KEY_MAP["BUTTON_L2"]=312
KEY_MAP["BUTTON_R2"]=313
KEY_MAP["BUTTON_SELECT"]=314
KEY_MAP["BUTTON_START"]=315
KEY_MAP["BUTTON_MODE"]=316
KEY_MAP["BUTTON_THUMBL"]=317
KEY_MAP["BUTTON_THUMBR"]=318
KEY_MAP["DPAD_CENTER"]=353
KEY_MAP["GUIDE"]=362
KEY_MAP["DVR"]=366
KEY_MAP["TV"]=377
KEY_MAP["CALENDAR"]=397
KEY_MAP["CHANNEL_UP"]=402
KEY_MAP["CHANNEL_DOWN"]=403
KEY_MAP["CONTACTS"]=429
KEY_MAP["FUNCTION"]=464
KEY_MAP["ESCAPE"]=465
KEY_MAP["F1"]=466
KEY_MAP["F2"]=467
KEY_MAP["F3"]=468
KEY_MAP["F4"]=469
KEY_MAP["F5"]=470
KEY_MAP["F6"]=471
KEY_MAP["F7"]=472
KEY_MAP["F8"]=473
KEY_MAP["F9"]=474
KEY_MAP["F10"]=475
KEY_MAP["F11"]=476
KEY_MAP["F12"]=477
KEY_MAP["1"]=478
KEY_MAP["2"]=479
KEY_MAP["D"]=480
KEY_MAP["E"]=481
KEY_MAP["F"]=482
KEY_MAP["S"]=483
KEY_MAP["B"]=484


# coding: utf-8

TAG_FUNCTION = 'func'
TAG_MEMORY = 'memory'
TAG_SYS_MEMORY = 'sys_memory'
TAG_BATTERY = 'battery'
TAG_CPU = 'cpu'
TAG_SNAPSHOT = 'screen'

AIRNATIVE = '/data/local/tmp/air-native'

ROTATION_0 = 0
ROTATION_90 = 1
ROTATION_180 = 2
ROTATION_270 = 3

ENV_SERIALNO = 'AIRTEST_SERIALNO'


#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
simple json log
'''
import json
import threading
import time

from airtest import base

class Lock(object):
    def __init__(self, tlock):
        self.lock = tlock
    
    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, *args):
        self.lock.release()

class JSONLog(object):
    def __init__(self, filename):
        self._filename = filename
        self._fd = None
        self._lock = threading.Lock()

    def writeline(self, d, *args):
        '''
        @param d (dict or string): content needed write to file
        @param args(array): only when d is string, support writeline('hello %s', 'world')
        '''
        with Lock(self._lock) as _:
            base.makedirs(base.dirname(self._filename))
            with open(self._filename, 'a') as file:
                if isinstance(d, dict):
                    d.update({'timestamp': int(time.time())})
                    outline = json.dumps(d)
                else:
                    outline = str(d) % args
                file.write(outline.rstrip() + '\n')


if __name__ == '__main__':
    log = JSONLog('test.log')
    log.writeline('hello')
    log.writeline({'hello': 'world'})



#!/usr/bin/env python
# coding: utf-8
#
# Licensed under MIT
#

import setuptools
setuptools.setup(setup_requires=['pbr'], pbr=True)

'''
# import sys
# sys.path.insert(1, '/usr/lib/pyshared/python2.7')
from setuptools import setup, find_packages

from airtest import __version__

long_description = ''
try:
    with open('README.md') as f:
        long_description = f.read()
except:
    pass

setup(name='airtest',
  version=__version__,
  description='mobile test(black air) python lib',
  long_description=long_description,

  author='codeskyblue',
  author_email='codeskyblue@gmail.com',

  packages=find_packages(),
  include_package_data=True,
  package_data={},
  install_requires=[
      #'Appium-Python-Client >= 0.10',
      'click >= 3.3',
      'fuckit >= 4.8.0',
      'humanize >= 0.5',
      'pystache >= 0.5.4',
      'aircv >= 1.02',
      'Flask >= 0.10.1',
      # 'paramiko',
      #'androidviewclient >= 7.1.1', 
      # 'requests >= 2.4.3', 
  '''#], entry_points='''
  #[console_scripts]
  #air.test = airtest.console:main
  #''')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import collections
import os
import platform
import time
import threading
import json

import cv2
import aircv as ac

from . import base
from . import proto
from . import patch
from .image import sift as imtsift
from .image import template as imttemplate

log = base.getLogger('devsuit')

class DeviceSuit(object):
    def __init__(self, devtype, dev, logfile='log/airtest.log'):
        # print 'DEVSUIT_SERIALNO:', phoneno
        self.dev = dev
        # self.appname = appname
        self._devtype = devtype
        self._inside_depth = 0

        # default image search extentension and 
        self._image_exts = ['.jpg', '.png']
        self._image_dirs = ['.', 'image']

        self._rotation = None # 0,1,2,3
        self._tmpdir = 'tmp'
        self._click_timeout = 20.0 # if icon not found in this time, then panic
        self._delay_after_click = 0.5 # when finished click, wait time
        self._screen_resolution = None

        self._snapshot_file = None
        self._keep_capture = False # for func:keepScreen,releaseScreen
        self._logfile = logfile
        self._loglock = threading.Lock()
        self._operation_mark = False

        self._image_match_method = 'auto'
        self._threshold = 0.3 # for findImage

        if self._logfile:
            logdir = os.path.dirname(logfile) or '.'
            if not os.path.exists(logdir):
                os.makedirs(logdir)
            if os.path.exists(logfile):
                backfile = logfile+'.'+time.strftime('%Y%m%d%H%M%S')
                os.rename(logfile, backfile)

        # Only for android phone method=<adb|screencap>
        def _snapshot_method(method):
            if method and self._devtype == 'android':
                self.dev._snapshot_method = method

        self._snapshot_method = _snapshot_method
        #-- end of func setting

    def __getattribute__(self, name):
        # print name
        v = object.__getattribute__(self, name)
        if isinstance(v, collections.Callable):
            objdict = object.__getattribute__(self, '__dict__')
            # print objdict

            def _wrapper(*args, **kwargs):
                objdict['_inside_depth'] += 1
                # log function call
                ret = v(*args, **kwargs)
                if objdict['_inside_depth'] == 1 and \
                    not v.__name__.startswith('_') and \
                        not v.__name__ == 'log':
                    self.log(proto.TAG_FUNCTION, dict(name=v.__name__, args=args, kwargs=kwargs))
                objdict['_inside_depth'] -= 1
                return ret
            return _wrapper
        return v

    def _imfind(self, bgimg, search):
        method = self._image_match_method
        print 'match-method:', method
        imsrc, imsch = ac.imread(bgimg), ac.imread(search)
        if method == 'auto':
            point = ac.find(imsrc, imsch)
        elif method == 'template':
            res = ac.find_template(imsrc, imsch, self._threshold)
            if res:
                point, score = res
                print 'match result:', point, score
                return point
            return None

        elif method == 'sift':
            point = imtsift.find(search, bgimg)
        else:
            raise RuntimeError("Unknown image match method: %s" %(method))
        return point

    def _imfindall(self, bgimg, search, maxcnt, sort):
        if not maxcnt:
            maxcnt = 0
        method = self._image_match_method
        imsrc, imsch = ac.imread(bgimg), ac.imread(search)
        if method == 'auto':
            points = ac.find_all(imsrc, imsch, maxcnt=5)
            # points = imtauto.locate_more_image_Template(search, bgimg, num=maxcnt)
        elif method == 'template':
            points = imttemplate.findall(search, bgimg, self._threshold, maxcnt=maxcnt)
        elif method == 'sift':
            points = imtsift.findall(search, bgimg, maxcnt=maxcnt)
        else:
            raise RuntimeError("Unknown image match method: %s" %(method))
        if sort:
            def cmpy((x0, y0), (x1, y1)):
                return y1<y0
            def cmpx((x0, y0), (x1, y1)):
                return x1<x1
            m = {'x': cmpx, 'y': cmpy}
            points.sort(cmp=m[sort])
        return points

    def rotation(self):
        '''
        device orientation
        @return int
        '''
        # 通过globalSet设置的rotation
        if self._rotation:
            return self._rotation
        # 看dev是否有rotation方法
        if hasattr(self.dev, 'rotation'):
            return self.dev.rotation()
        # windows上的特殊处理
        if self._devtype == 'windows':
            return proto.ROTATION_0
        return proto.ROTATION_0

    def _fixPoint(self, (x, y)):
        w, h = self.shape() # in shape() the width always < height
        if self.rotation() % 2 == 1:
            w, h = h, w

        if isinstance(x, float) and x <= 1.0:
            x = int(w*x)
        if isinstance(y, float) and y <= 1.0:
            y = int(h*y)
        return (x, y)

    def _searchImage(self, filename):
        ''' Search image in default path '''
        if isinstance(filename, unicode) and platform.system() == 'Windows':
            filename = filename.encode('gbk')
            #filename = filename.encode('utf-8')
        basename, ext = os.path.splitext(filename)
        exts = [ext] if ext else self._image_exts
        for folder in self._image_dirs:
            for ext in exts:
                fullpath = os.path.join(folder, basename+ext)
                if os.path.exists(fullpath):
                    return fullpath
        raise RuntimeError('Image file(%s) not found in %s' %(filename, self._image_dirs))

    def _PS2Point(self, PS):
        '''
        Convert PS to point
        @return (x, y) or None if not found
        '''
        if isinstance(PS, basestring):
            PS = self.find(PS)
            if not PS:
                return None
        (x, y) = self._fixPoint(PS)#(PS[0], PS[1]))#(1L, 2L))
        return (x, y)

    def _saveScreen(self, filename, random_name=True, tempdir=True):
        # use last snapshot file
        if self._snapshot_file and self._keep_capture:
            return self._snapshot_file

        if random_name:
            filename = base.random_name(filename)
        if tempdir:
            filename = os.path.join(self._tmpdir, filename)

        parent_dir = os.path.dirname(filename) or '.'
        if not os.path.exists(parent_dir):
            base.makedirs(parent_dir)

        # FIXME(ssx): don't save as file, better store in memory
        self.dev.snapshot(filename)

        if tempdir:
            self.log(proto.TAG_SNAPSHOT, dict(filename=filename))
        self._snapshot_file = filename
        return filename

    def log(self, tag, message):
        if not self._logfile:
            return

        self._loglock.acquire()
        timestamp = time.time()
        try:
            dirname = os.path.dirname(self._logfile) or '.'
            if not os.path.exists(dirname):
                os.path.makedirs(dirname)
        except:
            pass
        with open(self._logfile, 'a') as file:
            data = dict(timestamp=int(timestamp), tag=tag, data=message)
            file.write(json.dumps(data) + '\n')
        self._loglock.release()

    def keepCapture(self):
        '''
        Use screen in memory
        '''
        self._keep_capture = True

    def releaseCapture(self):
        '''
        Donot use screen in memory (this is default behavior)
        '''
        self._keep_capture = False

    def takeSnapshot(self, filename):
        '''
        Take screen snapshot

        @param filename: string (base filename want to save as basename)
        @return string: (filename that really save to)
        '''
        savefile = self._saveScreen(filename, random_name=False, tempdir=False)
        return savefile

    def globalSet(self, *args, **kwargs):
        '''
        app setting, be careful you should known what you are doing.
        @parma m(dict): eg:{"threshold": 0.3}
        '''
        if len(args) > 0:
            m = args[0]
            assert isinstance(m, dict)
        else:
            m = kwargs
        for k,v in m.items():
            key = '_'+k
            if hasattr(self, key):
                item = getattr(self, key)
                if callable(item):
                    item(v)
                else:
                    setattr(self, key, v)
            else:
                print 'not have such setting: %s' %(k)

    def globalGet(self, key):
        '''
        get app setting
        '''
        if hasattr(self, '_'+key):
            return getattr(self, '_'+key)
        return None

    def startApp(self, appname, activity):
        '''
        Start app
        '''
        self.dev.start_app(appname, activity)

    def stopApp(self, appname):
        '''
        Stop app
        '''
        self.dev.stop_app(appname)

    def find(self, imgfile):
        '''
        Find image position on screen

        @return (point founded or None if not found)
        '''
        filepath = self._searchImage(imgfile)
        
        log.debug('Locate image path: %s', filepath)
        
        screen = self._saveScreen('screen-{t}-XXXX.png'.format(t=time.strftime("%y%m%d%H%M%S")))
        if self._screen_resolution:
            # resize image
            ow, oh = self._screen_resolution # original
            cw, ch = self.shape() # current
            (ratew, rateh) = cw/float(ow), ch/float(oh)

            im = cv2.imread(filepath, cv2.IMREAD_UNCHANGED)

            nim = cv2.resize(im, (0, 0), fx=ratew, fy=rateh)
            new_name = base.random_name('resize-{t}-XXXX.png'.format(t=time.strftime("%y%m%d%H%M%S")))
            filepath = new_name = os.path.join(self._tmpdir, new_name)
            cv2.imwrite(new_name, nim)
            # im.resize((int(ratew*rw), int(rateh*rh))).save(new_name)
            # filepath = new_name
        pt = self._imfind(screen, filepath)
        return pt

    def mustFind(self, imgfile):
        ''' 
        Raise Error if image not found
        '''
        pt = self.find(imgfile)
        if not pt:
            raise RuntimeError("Image[%s] not found" %(imgfile))
        return pt

    def findall(self, imgfile, maxcnt=None, sort=None):
        '''
        Find multi positions that imgfile on screen

        @maxcnt (int): max number of object restricted.
        @sort (string): (None|x|y) x to sort with x, small in front, None to be origin order
        @return list point that found
        @warn not finished yet.
        '''
        filepath = self._searchImage(imgfile)
        screen = self._saveScreen('find-XXXXXXXX.png')
        pts = self._imfindall(screen, filepath, maxcnt, sort)
        return pts

    def safeWait(self, imgfile, seconds=20.0):
        '''
        Like wait, but don't raise RuntimeError

        return None when timeout
        return point if found
        '''
        try:
            return self.wait(imgfile, seconds)
        except:
            return None

    def wait(self, imgfile, timeout=20):
        '''
        Wait until some picture exists
        @return position when imgfile shows
        @raise RuntimeError if not found
        '''
        log.info('WAIT: %s', imgfile)
        start = time.time()
        while True:
            pt = self.find(imgfile)
            if pt:
                return pt
            if time.time()-start > timeout: 
                break
            time.sleep(1)
        raise RuntimeError('Wait timeout(%.2f)', float(timeout))

    def exists(self, imgfile):
        return True if self.find(imgfile) else False

    def click(self, SF, timeout=None, duration=None):
        '''
        Click function
        @param seconds: float (if time not exceed, it will retry and retry)
        '''
        if timeout == None:
            timeout = self._click_timeout
        log.info('CLICK %s, timeout=%.2fs, duration=%s', SF, timeout, str(duration))
        point = self._PS2Point(SF)
        if point:
            (x, y) = point
        else:
            (x, y) = self.wait(SF, timeout=timeout)
        log.info('Click %s point: (%d, %d)', SF, x, y)
        self.dev.touch(x, y, duration)
        log.debug('delay after click: %.2fs' ,self._delay_after_click)

        # FIXME(ssx): not tested
        if self._operation_mark:
            if self._snapshot_file and os.path.exists(self._snapshot_file):
                img = ac.imread(self._snapshot_file)
                ac.mark_point(img, (x, y))
                cv2.imwrite(self._snapshot_file, img)
            if self._devtype == 'android':
                self.dev.adbshell('am', 'broadcast', '-a', 'MP_POSITION', '--es', 'msg', '%d,%d'%(x, y))

        time.sleep(self._delay_after_click)

    def center(self):
        '''
        Center position
        '''
        w, h = self.shape()
        return w/2, h/2
    
    def clickIfExists(self, imgfile):
        '''
        Click when image file exists

        @return (True|False) if clicked
        '''
        log.info('CLICK IF EXISTS: %s' %(imgfile))
        pt = self.find(imgfile)
        if pt:
            log.debug('click for exists %s', imgfile)
            self.click(pt)
            return True
        else:
            log.debug('ignore for no exists %s', imgfile)
            return False

    def drag(self, fpt, tpt, duration=0.5):
        ''' 
        Drag from one place to another place

        @param fpt,tpt: filename or position
        @param duration: float (duration of the event in seconds)
        '''
        fpt = self._PS2Point(fpt)
        tpt = self._PS2Point(tpt)
        return self.dev.drag(fpt, tpt, duration)

    def sleep(self, secs=1.0):
        '''
        Sleeps for the specified number of seconds

        @param secs: float (number of seconds)
        @return None
        '''
        log.debug('SLEEP %.2fs', secs)
        time.sleep(secs)

    def type(self, text):
        '''
        Input some text

        @param text: string (text want to type)
        '''
        self.dev.type(text)

    @patch.run_once
    def shape(self):
        '''
        Get device shape

        @return (width, height), width < height
        '''
        return sorted(self.dev.shape())

    def keyevent(self, event):
        '''
        Send keyevent (only support android and ios)

        @param event: string (one of MENU,BACK,HOME)
        @return nothing
        '''
        if hasattr(self.dev, 'keyevent'):
            return self.dev.keyevent(event)
        raise RuntimeError('keyevent not support')





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import subprocess
import random
import string
import time
import logging
import threading


random.seed(time.time())
logging.basicConfig(format='%(asctime)s - %(levelname)s: %(message)s', level=logging.DEBUG) 

def getLogger(name='root'):
    return logging.getLogger(name)

log = getLogger('base')

def makedirs(dirname):
    try:
        os.makedirs(dirname)
    except:
        pass

def dirname(name):
    if os.path.isabs(name):
        return os.path.dirname(name)
    return os.path.dirname(os.path.abspath(name))

def exec_cmd(*cmds, **kwargs):
    '''
    @arguments env=None, timeout=3
    may raise Error
    '''
    env = os.environ.copy()
    env.update(kwargs.get('env', {}))
    envcopy = {}
    for key in env: 
        try:
            envcopy[key] = str(env[key]).encode('utf-8') # fix encoding
        except:
            print 'IGNORE BAD ENV KEY:', repr(key)
    env = envcopy

    timeout = kwargs.get('timeout', 120)
    shell = kwargs.get('shell', False)
    try:
        import sh
        log.debug('RUN(timeout=%ds): %s'%(timeout, ' '.join(cmds)))
        if shell:
            cmds = list(cmds)
            cmds[:0] = ['bash', '-c']
        c = sh.Command(cmds[0])
        try:
            r = c(*cmds[1:], _err_to_out=True, _out=sys.stdout, _env=env, _timeout=timeout)
        except:
            log.error('EXEC_CMD error, cmd: %s'%(' '.join(cmds)))
            raise
    except ImportError:
        log.debug('RUN(timeout=XX): %s'%(' '.join(cmds)))
        if shell:
            cmds = ' '.join(cmds)
        r = subprocess.Popen(cmds, env=env, stdout=sys.stdout, stderr=sys.stderr, shell=shell)
        return r.wait()
    return 0

def check_output(cmd):
    #log.debug('CHECK_OUTPUT, cmd: %s' %(cmd))
    return subprocess.check_output(cmd, shell=True)
    
def random_name(name):
    out = []
    for c in name:
        if c == 'X':
            c = random.choice(string.ascii_lowercase)
        out.append(c)
    return ''.join(out)


def wait_until(fn, interval=0.5, max_retry=10, args=(), kwargs={}):
    '''
    @return True(when found), False(when not found)
    '''
    log.debug('start wait func: %s', fn.__name__)
    retry = 0
    while retry < max_retry:
        retry += 1
        ret = fn(*args, **kwargs)
        if ret:
            return ret
        log.debug('wait until: %s, sleep: %s', fn.__name__, interval)
        time.sleep(interval)
    return None



# coding: utf-8

import uuid
import threading
import time
from . import patch

class Crontab(object):
    def __init__(self):
        self._tasks = {}
        self._running = False
        self._cycle = 3.0
        self._drain()

    def start(self):
        self._running = True

    def stop(self):
        self._running = False

    def running(self):
        return self._running

    def addfunc(self, func, *args, **kwargs):
        key = str(uuid.uuid4())
        self._tasks[key] = [func, args, kwargs]
        return key

    def delfunc(self, key):
        del(self._tasks[key])

    @patch.go
    def _drain(self):
        while True:
            if not self._running:
                time.sleep(1.5)
                continue
            for task in self._tasks.values():
                func, args, kwargs = task
                t = threading.Thread(target=func, args=args, kwargs=kwargs)
                t.daemon=True
                t.start()
            time.sleep(max(0.5, self._cycle))

if __name__ == '__main__':
    m = Crontab()
    def say(msg = 'hello'):
        print 'say:', msg
    m.addfunc(say, 'hi')
    key = m.addfunc(say, 'hey')
    m.start()
    time.sleep(2)
    m.delfunc(key)
    time.sleep(2)
    print 'stop', m.running()
    m.stop()
    time.sleep(2)

